'''
    MFEM Example 18 - Serial/Parallel Shared Code

    This is a python translation of ex18.hpp

    note: following variabls are set from ex18 or ex18p
        problem    
        num_equation
        max_char_speed 
        specific_heat_ratio;
        gas_constant;

'''
import numpy as np
from mfem import mfem_mode

if mfem_mode is None or mfem_mode == 'serial':
    import mfem.ser as mfem
else:
    import mfem.par as mfem

num_equation = 0
specific_heat_ratio = 0
gas_constant = 0
problem = 0
max_char_speed = 0


class FE_Evolution(mfem.TimeDependentOperator):
    def __init__(self, vfes, A, A_flux):
        self.dim = vfes.GetFE(0).GetDim()
        self.vfes = vfes
        self.A = A
        self.Aflux = A_flux
        self.Me_inv = mfem.DenseTensor(vfes.GetFE(0).GetDof(),
                                       vfes.GetFE(0).GetDof(),
                                       vfes.GetNE())

        self.state = mfem.Vector(num_equation)
        self.f = mfem.DenseMatrix(num_equation, self.dim)
        self.flux = mfem.DenseTensor(vfes.GetNDofs(), self.dim, num_equation)
        self.z = mfem.Vector(A.Height())

        dof = vfes.GetFE(0).GetDof()
        Me = mfem.DenseMatrix(dof)
        inv = mfem.DenseMatrixInverse(Me)
        mi = mfem.MassIntegrator()
        for i in range(vfes.GetNE()):
            mi.AssembleElementMatrix(vfes.GetFE(
                i), vfes.GetElementTransformation(i), Me)
            inv.Factor()
            inv.GetInverseMatrix(self.Me_inv(i))
        super(FE_Evolution, self).__init__(A.Height())

    def GetFlux(self, x, flux):
        state = self.state
        dof = self.flux.SizeI()
        dim = self.flux.SizeJ()

        flux_data = []
        for i in range(dof):
            for k in range(num_equation):
                self.state[k] = x[i, k]
            ComputeFlux(state, dim, self.f)

            flux_data.append(self.f.GetDataArray().transpose().copy())
            # flux[i].Print()
            # print(self.f.GetDataArray())
            # for d in range(dim):
            #    for k in range(num_equation):
            #        flux[i, d, k] = self.f[k, d]

            mcs = ComputeMaxCharSpeed(state, dim)
            if (mcs > globals()['max_char_speed']):
                globals()['max_char_speed'] = mcs

        flux.Assign(np.stack(flux_data))
        #print("max char speed", globals()['max_char_speed'])

    def Mult(self, x, y):
        globals()['max_char_speed'] = 0.
        num_equation = globals()['num_equation']
        # 1. Create the vector z with the face terms -<F.n(u), [w]>.
        self.A.Mult(x, self.z)

        # 2. Add the element terms.
        # i.  computing the flux approximately as a grid function by interpolating
        #     at the solution nodes.
        # ii. multiplying this grid function by a (constant) mixed bilinear form for
        #     each of the num_equation, computing (F(u), grad(w)) for each equation.

        xmat = mfem.DenseMatrix(
            x.GetData(), self.vfes.GetNDofs(), num_equation)
        self.GetFlux(xmat, self.flux)

        for k in range(num_equation):
            fk = mfem.Vector(self.flux[k].GetData(),
                             self.dim * self.vfes.GetNDofs())
            o = k * self.vfes.GetNDofs()
            zk = self.z[o: o+self.vfes.GetNDofs()]
            self.Aflux.AddMult(fk, zk)

        # 3. Multiply element-wise by the inverse mass matrices.
        zval = mfem.Vector()
        vdofs = mfem.intArray()
        dof = self.vfes.GetFE(0).GetDof()
        zmat = mfem.DenseMatrix()
        ymat = mfem.DenseMatrix(dof, num_equation)

        for i in range(self.vfes.GetNE()):
            # Return the vdofs ordered byNODES
            vdofs = mfem.intArray(self.vfes.GetElementVDofs(i))
            self.z.GetSubVector(vdofs, zval)
            zmat.UseExternalData(zval.GetData(), dof, num_equation)
            mfem.Mult(self.Me_inv[i], zmat, ymat)
            y.SetSubVector(vdofs, ymat.GetData())


class DomainIntegrator(mfem.BilinearFormIntegrator):
    def __init__(self, dim):
        num_equation = globals()['num_equation']
        self.flux = mfem.DenseMatrix(num_equation, dim)
        self.shape = mfem.Vector()
        self.dshapedr = mfem.DenseMatrix()
        self.dshapedx = mfem.DenseMatrix()
        super(DomainIntegrator, self).__init__()

    def AssembleElementMatrix2(self, trial_fe, test_fe, Tr, elmat):
        # Assemble the form (vec(v), grad(w))

        # Trial space = vector L2 space (mesh dim)
        # Test space  = scalar L2 space

        dof_trial = trial_fe.GetDof()
        dof_test = test_fe.GetDof()
        dim = trial_fe.GetDim()

        self.shape.SetSize(dof_trial)
        self.dshapedr.SetSize(dof_test, dim)
        self.dshapedx.SetSize(dof_test, dim)

        elmat.SetSize(dof_test, dof_trial * dim)
        elmat.Assign(0.0)

        maxorder = max(trial_fe.GetOrder(), test_fe.GetOrder())
        intorder = 2 * maxorder
        ir = mfem.IntRules.Get(trial_fe.GetGeomType(), intorder)

        for i in range(ir.GetNPoints()):
            ip = ir.IntPoint(i)

            # Calculate the shape functions
            trial_fe.CalcShape(ip, self.shape)
            self.shape *= ip.weight

            # Compute the physical gradients of the test functions
            Tr.SetIntPoint(ip)
            test_fe.CalcDShape(ip, self.dshapedr)
            mfem.Mult(self.dshapedr, Tr.AdjugateJacobian(), self.dshapedx)

            for d in range(dim):
                for j in range(dof_test):
                    for k in range(dof_trial):
                        elmat[j, k + d * dof_trial] += self.shape[k] * \
                            self.dshapedx[j, d]


class FaceIntegrator(mfem.NonlinearFormIntegrator):
    def __init__(self, rsolver, dim):
        self.rsolver = rsolver
        self.shape1 = mfem.Vector()
        self.shape2 = mfem.Vector()
        self.funval1 = mfem.Vector(num_equation)
        self.funval2 = mfem.Vector(num_equation)
        self.nor = mfem.Vector(dim)
        self.fluxN = mfem.Vector(num_equation)
        self.eip1 = mfem.IntegrationPoint()
        self.eip2 = mfem.IntegrationPoint()
        super(FaceIntegrator, self).__init__()

        self.fluxNA = np.atleast_2d(self.fluxN.GetDataArray())

    def AssembleFaceVector(self, el1, el2, Tr, elfun, elvect):
        num_equation = globals()['num_equation']
        # Compute the term <F.n(u),[w]> on the interior faces.
        dof1 = el1.GetDof()
        dof2 = el2.GetDof()

        self.shape1.SetSize(dof1)
        self.shape2.SetSize(dof2)

        elvect.SetSize((dof1 + dof2) * num_equation)
        elvect.Assign(0.0)

        elfun1_mat = mfem.DenseMatrix(elfun.GetData(), dof1, num_equation)
        elfun2_mat = mfem.DenseMatrix(
            elfun[dof1*num_equation:].GetData(), dof2, num_equation)

        elvect1_mat = mfem.DenseMatrix(elvect.GetData(), dof1, num_equation)
        elvect2_mat = mfem.DenseMatrix(
            elvect[dof1*num_equation:].GetData(), dof2, num_equation)

        # Integration order calculation from DGTraceIntegrator
        if (Tr.Elem2No >= 0):
            intorder = (min(Tr.Elem1.OrderW(), Tr.Elem2.OrderW()) +
                        2*max(el1.GetOrder(), el2.GetOrder()))
        else:
            intorder = Tr.Elem1.OrderW() + 2*el1.GetOrder()

        if (el1.Space() == mfem.FunctionSpace().Pk):
            intorder += 1

        ir = mfem.IntRules.Get(Tr.GetGeometryType(), int(intorder))

        mat1A = elvect1_mat.GetDataArray()
        mat2A = elvect2_mat.GetDataArray()
        shape1A = np.atleast_2d(self.shape1.GetDataArray())
        shape2A = np.atleast_2d(self.shape2.GetDataArray())

        for i in range(ir.GetNPoints()):
            ip = ir.IntPoint(i)
            Tr.Loc1.Transform(ip, self.eip1)
            Tr.Loc2.Transform(ip, self.eip2)

            # Calculate basis functions on both elements at the face
            el1.CalcShape(self.eip1, self.shape1)
            el2.CalcShape(self.eip2, self.shape2)

            # Interpolate elfun at the point
            elfun1_mat.MultTranspose(self.shape1, self.funval1)
            elfun2_mat.MultTranspose(self.shape2, self.funval2)
            Tr.Face.SetIntPoint(ip)

            # Get the normal vector and the flux on the face

            mfem.CalcOrtho(Tr.Face.Jacobian(), self.nor)

            mcs = self.rsolver.Eval(
                self.funval1, self.funval2, self.nor, self.fluxN)

            # Update max char speed
            if mcs > globals()['max_char_speed']:
                globals()['max_char_speed'] = mcs

            self.fluxN *= ip.weight

            #
            mat1A -= shape1A.transpose().dot(self.fluxNA)
            mat2A += shape2A.transpose().dot(self.fluxNA)
            '''
            for k in range(num_equation):
                for s in range(dof1):
                    elvect1_mat[s, k] -= self.fluxN[k] * self.shape1[s]
                for s in range(dof2):                   
                    elvect2_mat[s, k] += self.fluxN[k] * self.shape2[s]
            '''


class RiemannSolver(object):
    def __init__(self):
        num_equation = globals()['num_equation']
        self.flux1 = mfem.Vector(num_equation)
        self.flux2 = mfem.Vector(num_equation)

    def Eval(self, state1, state2, nor, flux):

        # NOTE: nor in general is not a unit normal
        dim = nor.Size()

        assert StateIsPhysical(state1, dim), ""
        assert StateIsPhysical(state2, dim), ""

        maxE1 = ComputeMaxCharSpeed(state1, dim)
        maxE2 = ComputeMaxCharSpeed(state2, dim)
        maxE = max(maxE1, maxE2)

        ComputeFluxDotN(state1, nor, self.flux1)
        ComputeFluxDotN(state2, nor, self.flux2)

        #normag = np.sqrt(np.sum(nor.GetDataArray()**2))
        normag = nor.Norml2()

        '''
        for i in range(num_equation):
            flux[i] = (0.5 * (self.flux1[i] + self.flux2[i])
                       - 0.5 * maxE * (state2[i] - state1[i]) * normag)
        '''
        f = (0.5 * (self.flux1.GetDataArray() + self.flux2.GetDataArray())
             - 0.5 * maxE * (state2.GetDataArray() - state1.GetDataArray()) * normag)
        flux.Assign(f)

        return maxE


def StateIsPhysical(state, dim):
    specific_heat_ratio = globals()["specific_heat_ratio"]

    den = state[0]
    #den_vel = state.GetDataArray()[1:1+dim]
    den_energy = state[1 + dim]

    if (den < 0):
        print("Negative density: " + str(state.GetDataArray()))
        return False
    if (den_energy <= 0):
        print("Negative energy: " + str(state.GetDataArray()))
        return False

    #den_vel2 = np.sum(den_vel**2)/den
    den_vel2 = (state[1:1+dim].Norml2())**2/den
    pres = (specific_heat_ratio - 1.0) * (den_energy - 0.5 * den_vel2)
    if (pres <= 0):
        print("Negative pressure: " + str(state.GetDataArray()))
        return False
    return True


class InitialCondition(mfem.VectorPyCoefficient):
    def __init__(self, dim):
        mfem.VectorPyCoefficient.__init__(self, dim)

    def EvalValue(self, x):
        dim = x.shape[0]
        assert dim == 2, ""
        problem = globals()['problem']
        if (problem == 1):
            # "Fast vortex"
            radius = 0.2
            Minf = 0.5
            beta = 1. / 5.
        elif (problem == 2):
            # "Slow vortex"
            radius = 0.2
            Minf = 0.05
            beta = 1. / 50.
        else:
            assert False, "Cannot recognize problem. Options are: 1 - fast vortex, 2 - slow vortex"

        xc = 0.0
        yc = 0.0
        # Nice units
        vel_inf = 1.
        den_inf = 1.

        specific_heat_ratio = globals()["specific_heat_ratio"]
        gas_constant = globals()["gas_constant"]

        pres_inf = (den_inf / specific_heat_ratio) * \
            (vel_inf / Minf) * (vel_inf / Minf)
        temp_inf = pres_inf / (den_inf * gas_constant)

        r2rad = 0.0
        r2rad += (x[0] - xc) * (x[0] - xc)
        r2rad += (x[1] - yc) * (x[1] - yc)
        r2rad /= (radius * radius)

        shrinv1 = 1.0 / (specific_heat_ratio - 1.)

        velX = vel_inf * \
            (1 - beta * (x[1] - yc) / radius * np.exp(-0.5 * r2rad))
        velY = vel_inf * beta * (x[0] - xc) / radius * np.exp(-0.5 * r2rad)
        vel2 = velX * velX + velY * velY

        specific_heat = gas_constant * specific_heat_ratio * shrinv1

        temp = temp_inf - (0.5 * (vel_inf * beta) *
                           (vel_inf * beta) / specific_heat * np.exp(-r2rad))

        den = den_inf * (temp/temp_inf)**shrinv1
        pres = den * gas_constant * temp
        energy = shrinv1 * pres / den + 0.5 * vel2

        y = np.array([den, den * velX, den * velY, den * energy])
        return y


def ComputePressure(state, dim):
    den = state[0]
    #den_vel = state.GetDataArray()[1:1+dim]
    den_energy = state[1 + dim]

    specific_heat_ratio = globals()["specific_heat_ratio"]
    #den_vel2 = np.sum(den_vel**2)/den
    den_vel2 = (state[1:1+dim].Norml2())**2/den
    pres = (specific_heat_ratio - 1.0) * (den_energy - 0.5 * den_vel2)

    return pres


def ComputeFlux(state, dim, flux):
    den = state[0]
    den_vel = state.GetDataArray()[1:1+dim]
    den_energy = state[1 + dim]

    assert StateIsPhysical(state, dim), ""

    pres = ComputePressure(state, dim)

    den_vel2 = np.atleast_2d(den_vel)
    fluxA = flux.GetDataArray()
    fluxA[0, :] = den_vel
    fluxA[1:1+dim, :] = den_vel2.transpose().dot(den_vel2) / den
    for d in range(dim):
        fluxA[1+d, d] += pres

    H = (den_energy + pres) / den
    flux.GetDataArray()[1+dim, :] = den_vel * H


def ComputeFluxDotN(state, nor, fluxN):
    # NOTE: nor in general is not a unit normal
    dim = nor.Size()
    nor = nor.GetDataArray()
    fluxN = fluxN.GetDataArray()

    den = state[0]
    den_vel = state.GetDataArray()[1:1+dim]
    den_energy = state[1 + dim]

    assert StateIsPhysical(state, dim), ""

    pres = ComputePressure(state, dim)

    den_velN = den_vel.dot(nor)

    fluxN[0] = den_velN
    fluxN[1:1+dim] = den_velN * den_vel / den + pres * nor

    H = (den_energy + pres) / den
    fluxN[1+dim] = den_velN * H


def ComputeMaxCharSpeed(state, dim):
    specific_heat_ratio = globals()["specific_heat_ratio"]

    den = state[0]
    den_vel2 = (state[1:1+dim].Norml2())**2/den
    pres = ComputePressure(state, dim)

    sound = np.sqrt(specific_heat_ratio * pres / den)
    vel = np.sqrt(den_vel2 / den)

    return vel + sound



'''
 This example shows
   constructing HYPREParCSR in a distributed manner
     1) build scipy.sparse on each node
           each node has informaiton of small number of rows
     2) assemble it to HYPREParCSR
           scipy matries are concatenated vertical

   convert back ParCSR to scipy.sparse
     1) on each node, it produce scipy.sparse which contain
        only its own rows.


   example_4 tests
      how to make complex matrix
      transpose/conj of complex matrix

      mpirun -np 6 python2.7 example_4.py
'''
import mfem.par as par
from mfem.common.chypre import *

from scipy.sparse import csr_matrix, coo_matrix

from mpi4py import MPI
comm     = MPI.COMM_WORLD     
num_proc = MPI.COMM_WORLD.size
myid     = MPI.COMM_WORLD.rank

def print_hypre(M, txt):
    for i in range(num_proc):
       MPI.COMM_WORLD.Barrier()                              
       if myid == i:
          if myid == 0:
              print(txt)
              print('MyID: ', myid)
          else:
              print('MyID: ', myid)
          m  = ToScipyCoo(M)              
          print('shape = ', m.shape)
          print(m.toarray())
       MPI.COMM_WORLD.Barrier()                              
shape = (1, 6)
# make sample matrix
row  = np.array([0, 0, ])
col  = np.array([0, 3, ])
data = np.array([1, 3.*(myid+1), ])
m = coo_matrix((data, (row, col)), shape=shape)
m = m.tocsr()


M1 = CHypreMat(m, m)
M2 = M1.transpose().conj()

print_hypre(M1[1], 'matrix M1 (imag)')
print_hypre(M2[1], 'matrix M2 (imag)')








'''
   MFEM example 4p

   See c++ version in the MFEM library for more detail 

   How to run:
      mpirun -np 2 python <arguments>

   Example of arguments:
      ex4p.py -m square-disc.mesh
      ex4p.py -m star.mesh
      ex4p.py -m beam-tet.mesh
      ex4p.py -m beam-hex.mesh
      ex4p.py -m escher.mesh -o 2 -sc
      ex4p.py -m fichera.mesh -o 2 -hb
      ex4p.py -m fichera-q2.vtk
      ex4p.py -m fichera-q3.mesh -o 2 -sc
      ex4p.py -m square-disc-nurbs.mesh -o 3
      ex4p.py -m beam-hex-nurbs.mesh -o 3
      ex4p.py -m periodic-square.mesh -no-bc
      ex4p.py -m periodic-cube.mesh -no-bc
      ex4p.py -m amr-quad.mesh
      ex4p.py -m amr-hex.mesh -o 2 -sc
      ex4p.py -m amr-hex.mesh -o 2 -hb
      ex4p.py -m star-surf.mesh -o 3 -hb
'''
from numpy import sin, array, cos
import numpy as np
from os.path import expanduser, join, dirname
import sys
from mfem.common.arg_parser import ArgParser
import mfem.par as mfem
from mpi4py import MPI
num_procs = MPI.COMM_WORLD.size
myid = MPI.COMM_WORLD.rank


parser = ArgParser(description='Ex4 ')
parser.add_argument('-m', '--mesh',
                    default='star.mesh',
                    action='store', type=str,
                    help='Mesh file to use.')
parser.add_argument('-o', '--order',
                    action='store', default=1, type=int,
                    help="Finite element order (polynomial degree)")
parser.add_argument('-no-bc', '--dont-impose-bc',
                    action='store_false',  default=True,
                    help="Impose or not essential boundary conditions.")
parser.add_argument("-f", "--frequency",
                    action='store', default=1.0, type=float,
                    help="Set the frequency for the exact solution.")
parser.add_argument("-sc", "--static-condensation",
                    action='store_true', default=False,
                    help="Enable static condensation.")
parser.add_argument("-hb", "--hybridization",
                    action='store_true', default=False,
                    help="Enable hybridization.")
parser.add_argument('-vis', '--visualization',
                    action='store_true',
                    help='Enable GLVis visualization')
args = parser.parse_args()

order = args.order
set_bc = args.dont_impose_bc
static_cond = args.static_condensation
hybridization = args.hybridization
freq = args.frequency
kappa = np.pi*freq
visualization = args.visualization
if (myid == 0):
    parser.print_options(args)

device = mfem.Device('cpu')
if myid == 0:
    device.Print()

# 3. Read the (serial) mesh from the given mesh file on all processors.  We
#    can handle triangular, quadrilateral, tetrahedral, hexahedral, surface
#    and volume, as well as periodic meshes with the same code.
meshfile = expanduser(join(dirname(__file__), '..', 'data', args.mesh))
mesh = mfem.Mesh(meshfile, 1, 1)
dim = mesh.Dimension()
sdim = mesh.SpaceDimension()


#  Coefficients
class E_exact(mfem.VectorPyCoefficient):
    def __init__(self, sdim):
        mfem.VectorPyCoefficient.__init__(self, sdim)

    def EvalValue(self, p):
        dim = p.shape[0]
        x = p[0]
        y = p[1]
        F0 = cos(kappa*x)*sin(kappa*y)
        F1 = cos(kappa*y)*sin(kappa*x)
        if dim == 3:
            return (F0, F1, 0.0)
        else:
            return (F0, F1)


class f_exact(mfem.VectorPyCoefficient):
    def __init__(self, sdim):
        mfem.VectorPyCoefficient.__init__(self, sdim)

    def EvalValue(self, p):
        dim = p.shape[0]
        x = p[0]
        y = p[1]
        temp = 1 + 2*kappa*kappa

        F0 = temp * cos(kappa*x)*sin(kappa*y)
        F1 = temp * cos(kappa*y)*sin(kappa*x)
        if dim == 3:
            return (F0, F1, 0.0)
        else:
            return (F0, F1)


# 4. Refine the serial mesh on all processors to increase the resolution. In
#    this example we do 'ref_levels' of uniform refinement. We choose
#    'ref_levels' to be the largest number that gives a final mesh with no
#    more than 1,000 elements.
ref_levels = int(np.floor(np.log(1000./mesh.GetNE())/np.log(2.)/dim))
for x in range(ref_levels):
    mesh.UniformRefinement()

# 5. Define a parallel mesh by a partitioning of the serial mesh. Refine
#    this mesh further in parallel to increase the resolution. Once the
#    parallel mesh is defined, the serial mesh can be deleted. Tetrahedral
#    meshes need to be reoriented before we can define high-order Nedelec
#    spaces on them (this is needed in the ADS solver below).
pmesh = mfem.ParMesh(MPI.COMM_WORLD, mesh)
del mesh
par_ref_levels = 2
for l in range(par_ref_levels):
    pmesh.UniformRefinement()
pmesh.ReorientTetMesh()

# 6. Define a parallel finite element space on the parallel mesh. Here we
#    use the Raviart-Thomas finite elements of the specified order.
fec = mfem.RT_FECollection(order-1, dim)
fespace = mfem.ParFiniteElementSpace(pmesh, fec)
size = fespace.GlobalTrueVSize()
if myid == 0:
    print("Number of finite element unknows : " + str(size))

# 7. Determine the list of true (i.e. parallel conforming) essential
#    boundary dofs. In this example, the boundary conditions are defined
#    by marking all the boundary attributes from the mesh as essential
#    (Dirichlet) and converting them to a list of true dofs.
ess_tdof_list = mfem.intArray()
if pmesh.bdr_attributes.Size():
    ess_bdr = mfem.intArray(pmesh.bdr_attributes.Max())
    if set_bc:
        ess_bdr.Assign(1)
    else:
        ess_bdr.Assign(0)
    fespace.GetEssentialTrueDofs(ess_bdr, ess_tdof_list)

# 8. Set up the parallel linear form b(.) which corresponds to the
#    right-hand side of the FEM linear system, which in this case is
#    (f,phi_i) where f is given by the function f_exact and phi_i are the
#    basis functions in the finite element fespace.
f = f_exact(sdim)
b = mfem.ParLinearForm(fespace)
b.AddDomainIntegrator(mfem.VectorFEDomainLFIntegrator(f))
b.Assemble()

# 9. Define the solution vector x as a parallel finite element grid function
#    corresponding to fespace. Initialize x by projecting the exact
#    solution. Note that only values from the boundary faces will be used
#    when eliminating the non-homogeneous boundary condition to modify the
#    r.h.s. vector b.
x = mfem.ParGridFunction(fespace)
F = E_exact(sdim)
x.ProjectCoefficient(F)

# 10. Set up the parallel bilinear form corresponding to the H(div)
#     diffusion operator grad alpha div + beta I, by adding the div-div and
#     the mass domain integrators.

alpha = mfem.ConstantCoefficient(1.0)
beta = mfem.ConstantCoefficient(1.0)
a = mfem.ParBilinearForm(fespace)
a.AddDomainIntegrator(mfem.DivDivIntegrator(alpha))
a.AddDomainIntegrator(mfem.VectorFEMassIntegrator(beta))

# 11. Assemble the parallel bilinear form and the corresponding linear
#     system, applying any necessary transformations such as: parallel
#     assembly, eliminating boundary conditions, applying conforming
#     constraints for non-conforming AMR, static condensation,
#     hybridization, etc.
if (static_cond):
    a.EnableStaticCondensation()
elif (hybridization):
    hfec = mfem.DG_Interface_FECollection(order-1, dim)
    hfes = mfem.ParFiniteElementSpace(mesh, hfec)
    a.EnableHybridization(hfes, mfem.NormalTraceJumpIntegrator(),
                          ess_tdof_list)
a.Assemble()

A = mfem.HypreParMatrix()
B = mfem.Vector()
X = mfem.Vector()
a.FormLinearSystem(ess_tdof_list, x, b, A, X, B)

glob_size = A.GetGlobalNumRows()
if myid == 0:
    print("Size of linear system: " + str(glob_size))

# 12. Define and apply a parallel PCG solver for A X = B with the 2D AMS or
#     the 3D ADS preconditioners from hypre. If using hybridization, the
#     system is preconditioned with hypre's BoomerAMG.

pcg = mfem.CGSolver(A.GetComm())
pcg.SetOperator(A)
pcg.SetRelTol(1e-12)
pcg.SetMaxIter(500)
pcg.SetPrintLevel(1)
if hybridization:
    prec = mfem.HypreBoomerAMG(A)
else:
    if a.StaticCondensationIsEnabled():
        prec_fespace = a.SCParFESpace()
    else:
        prec_fespace = fespace
    if dim == 2:
        prec = mfem.HypreAMS(A, prec_fespace)
    else:
        prec = mfem.HypreADS(A, prec_fespace)

pcg.SetPreconditioner(prec)
pcg.Mult(B, X)

# 13. Recover the parallel grid function corresponding to X. This is the
#     local finite element solution on each processor.
a.RecoverFEMSolution(X, b, x)

# 14. Compute and print the L^2 norm of the error.
err = x.ComputeL2Error(F)
if myid == 0:
    print("|| F_h - F ||_{L^2} = " + str(err))

# 15. Save the refined mesh and the solution in parallel. This output can
#     be viewed later using GLVis: "glvis -np <np> -m mesh -g sol".
smyid = '{:0>6d}'.format(myid)
mesh_name = "mesh."+smyid
sol_name = "sol."+smyid
pmesh.Print(mesh_name, 8)
x.Save(sol_name, 8)

# 16. Send the solution by socket to a GLVis server.
if visualization:
    sol_sock = mfem.socketstream("localhost", 19916)
    sol_sock.send_text("parallel " + str(num_procs) + " " + str(myid))
    sol_sock.precision(8)
    sol_sock.send_solution(pmesh,   x)



'''
   MFEM example 29
      See c++ version in the MFEM library for more detail 
'''
import os
import mfem.ser as mfem
from mfem.ser import intArray
from os.path import expanduser, join, dirname
import numpy as np
from numpy import sin, cos, exp, sqrt, pi, abs, array


def GetMesh(mesh_type):
    if type == 3:
        mesh = mfem.Mesh(2, 12, 16, 8, 3)
        mesh.AddVertex(-1.0, -1.0, 0.0)
        mesh.AddVertex(1.0, -1.0, 0.0)
        mesh.AddVertex(1.0,  1.0, 0.0)
        mesh.AddVertex(-1.0,  1.0, 0.0)
        mesh.AddVertex(-1.0, -1.0, 1.0)
        mesh.AddVertex(1.0, -1.0, 1.0)
        mesh.AddVertex(1.0,  1.0, 1.0)
        mesh.AddVertex(-1.0,  1.0, 1.0)
        mesh.AddVertex(0.0, -1.0, 0.5)
        mesh.AddVertex(1.0,  0.0, 0.5)
        mesh.AddVertex(0.0,  1.0, 0.5)
        mesh.AddVertex(-1.0,  0.0, 0.5)

        mesh.AddTriangle(0, 1, 8)
        mesh.AddTriangle(1, 5, 8)
        mesh.AddTriangle(5, 4, 8)
        mesh.AddTriangle(4, 0, 8)
        mesh.AddTriangle(1, 2, 9)
        mesh.AddTriangle(2, 6, 9)
        mesh.AddTriangle(6, 5, 9)
        mesh.AddTriangle(5, 1, 9)
        mesh.AddTriangle(2, 3, 10)
        mesh.AddTriangle(3, 7, 10)
        mesh.AddTriangle(7, 6, 10)
        mesh.AddTriangle(6, 2, 10)
        mesh.AddTriangle(3, 0, 11)
        mesh.AddTriangle(0, 4, 11)
        mesh.AddTriangle(4, 7, 11)
        mesh.AddTriangle(7, 3, 11)

        mesh.AddBdrSegment(0, 1, 1)
        mesh.AddBdrSegment(1, 2, 1)
        mesh.AddBdrSegment(2, 3, 1)
        mesh.AddBdrSegment(3, 0, 1)
        mesh.AddBdrSegment(5, 4, 2)
        mesh.AddBdrSegment(6, 5, 2)
        mesh.AddBdrSegment(7, 6, 2)
        mesh.AddBdrSegment(4, 7, 2)
    elif mesh_type == 4:
        mesh = mfem.Mesh(2, 8, 4, 8, 3)

        mesh.AddVertex(-1.0, -1.0, 0.0)
        mesh.AddVertex(1.0, -1.0, 0.0)
        mesh.AddVertex(1.0,  1.0, 0.0)
        mesh.AddVertex(-1.0,  1.0, 0.0)
        mesh.AddVertex(-1.0, -1.0, 1.0)
        mesh.AddVertex(1.0, -1.0, 1.0)
        mesh.AddVertex(1.0,  1.0, 1.0)
        mesh.AddVertex(-1.0,  1.0, 1.0)

        mesh.AddQuad(0, 1, 5, 4)
        mesh.AddQuad(1, 2, 6, 5)
        mesh.AddQuad(2, 3, 7, 6)
        mesh.AddQuad(3, 0, 4, 7)

        mesh.AddBdrSegment(0, 1, 1)
        mesh.AddBdrSegment(1, 2, 1)
        mesh.AddBdrSegment(2, 3, 1)
        mesh.AddBdrSegment(3, 0, 1)
        mesh.AddBdrSegment(5, 4, 2)
        mesh.AddBdrSegment(6, 5, 2)
        mesh.AddBdrSegment(7, 6, 2)
        mesh.AddBdrSegment(4, 7, 2)

    else:
        assert False, "Unrecognized mesh type :" + str(mesh_type)

    mesh.FinalizeTopology()

    return mesh


def run(order=3,
        ref_levels=0,
        static_cond=False,
        mesh_type=4,
        mesh_order=3,
        visualization=True):

    # 2. Construct a quadrilateral or triangular mesh with the topology of a
    #    cylindrical surface.
    mesh = GetMesh(mesh_type)
    dim = mesh.Dimension()
    sdim = mesh.SpaceDimension()

    # 3. Refine the mesh to increase the resolution. In this example we do
    #   'ref_levels' of uniform refinement.
    for l in range(ref_levels):
        mesh.UniformRefinement()

    # 4. Transform the mesh so that it has a more interesting geometry.
    mesh.SetCurvature(mesh_order)

    class cTrans(mfem.VectorPyCoefficient):
        def __init__(self):
            mfem.VectorPyCoefficient.__init__(self, sdim)

        def EvalValue(self, x):
            tol = 1e-6
            theta = 0.0
            if abs(x[1] + 1.0) < tol:
                theta = 0.25 * pi * (x[0] - 2.0)
            elif abs(x[0] - 1.0) < tol:
                theta = 0.25 * pi * x[1]
            elif abs(x[1] - 1.0) < tol:
                theta = 0.25 * pi * (2.0 - x[0])
            elif abs(x[0] + 1.0) < tol:
                theta = 0.25 * pi * (4.0 - x[1])
            else:
                print("side not recognized " +
                      str(x[0]) + " " + str(x[1]) + " " + str(x[2]))

            return (cos(theta), sin(theta),
                    0.25 * (2.0 * x[2] - 1.0) * (cos(theta) + 2.0),)

    trans = cTrans()
    mesh.Transform(trans)

    # 5. Define a finite element space on the mesh. Here we use continuous
    #    Lagrange finite elements of the specified order.
    fec = mfem.H1_FECollection(order, dim)
    fespace = mfem.FiniteElementSpace(mesh, fec)
    print("Number of finite element unknowns: " + str(fespace.GetTrueVSize()))

    # 6. Determine the list of true (i.e. conforming) essential boundary dofs.
    #    In this example, the boundary conditions are defined by marking all
    #    the boundary attributes from the mesh as essential (Dirichlet) and
    #    converting them to a list of true dofs.
    ess_tdof_list = mfem.intArray()
    if mesh.bdr_attributes.Size():
        ess_bdr = mfem.intArray(mesh.bdr_attributes.Max())
        ess_bdr.Assign(1)
        fespace.GetEssentialTrueDofs(ess_bdr, ess_tdof_list)

    # 7. Set up the linear form b(.) which corresponds to the right-hand side of
    #    the FEM linear system, which in this case is (1,phi_i) where phi_i are
    #    the basis functions in the finite element fespace.
    b = mfem.LinearForm(fespace)
    one = mfem.ConstantCoefficient(1.0)
    b.AddDomainIntegrator(mfem.DomainLFIntegrator(one))
    b.Assemble()

    # 8. Define the solution vector x as a finite element grid function
    #    corresponding to fespace. Initialize x with initial guess of zero,
    #    which satisfies the boundary conditions.
    x = mfem.GridFunction(fespace)
    x.Assign(0.0)

    # 9. Set up the bilinear form a(.,.) on the finite element space
    #    corresponding to the Laplacian operator -Delta, by adding the Diffusion
    #    domain integrator.
    def sigmaFunc(x):
        a = 17.0 - 2.0 * x[0] * (1.0 + x[0])
        s = array([[0.5 + x[0] * x[0] * (8.0 / a - 0.5),
                    x[0] * x[1] * (8.0 / a - 0.5),
                    0.0],
                   [x[0] * x[1] * (8.0 / a - 0.5),
                    0.5 * x[0] * x[0] + 8.0 * x[1] * x[1] / a,
                    0.0],
                   [0.0, 0.0, a/32.]])
        return s

    class cSigma(mfem.MatrixPyCoefficient):
        def __init__(self):
            mfem.MatrixPyCoefficient.__init__(self, sdim)

        def EvalValue(self, x):
            return sigmaFunc(x)

    a = mfem.BilinearForm(fespace)
    sigma = cSigma()
    integ = mfem.DiffusionIntegrator(sigma)
    a.AddDomainIntegrator(integ)

    # 10. Assemble the bilinear form and the corresponding linear system,
    #     applying any necessary transformations such as: eliminating boundary
    #     conditions, applying conforming constraints for non-conforming AMR,
    #     static condensation, etc.
    if static_cond:
        a.EnableStaticCondensation()
    a.Assemble()

    A = mfem.OperatorPtr()
    B = mfem.Vector()
    X = mfem.Vector()

    a.FormLinearSystem(ess_tdof_list, x, b, A, X, B)

    print("Size of linear system: " + str(A.Height()))

    # 11. Solve the linear system A X = B.
    #     Use a simple symmetric Gauss-Seidel preconditioner with PCG.
    AA = mfem.OperatorHandle2SparseMatrix(A)
    M = mfem.GSSmoother(AA)
    mfem.PCG(A, M, B, X, 1, 200, 1e-12, 0.0)

    # 12. Recover the solution as a finite element grid function.
    a.RecoverFEMSolution(X, b, x)

    # 13. Compute error in the solution and its flux
    class cuExact(mfem.PyCoefficient):
        def __init__(self):
            mfem.PyCoefficient.__init__(self)

        def EvalValue(self, x):
            return (0.25 * (2.0 + x[0]) - x[2]) * (x[2] + 0.25 * (2.0 + x[0]))

    def duExactFunc(x):
        return (0.125 * (2.0 + x[0]) * x[1] * x[1],
                -0.125 * (2.0 + x[0]) * x[0] * x[1],
                -2.0 * x[2])

    class cduExact(mfem.VectorPyCoefficient):
        def __init__(self):
            mfem.VectorPyCoefficient.__init__(self, sdim)

        def EvalValue(self, x):
            return duExactFunc(x)

    class cfluxExact(mfem.VectorPyCoefficient):
        def __init__(self):
            mfem.VectorPyCoefficient.__init__(self, sdim)

        def EvalValue(self, x):
            s = sigmaFunc(x)
            du = duExactFunc(x)
            return -s.dot(du)

    uCoef = cuExact()
    err = x.ComputeL2Error(uCoef)

    print("|u - u_h|_2 = " + "{:g}".format(err))

    flux_fespace = mfem.FiniteElementSpace(mesh, fec, 3)
    flux = mfem.GridFunction(flux_fespace)
    x.ComputeFlux(integ, flux)
    flux *= -1.0

    fluxCoef = cfluxExact()
    flux_err = flux.ComputeL2Error(fluxCoef)

    print("|f - f_h|_2 = " + "{:g}".format(flux_err))

    # 14. Save the refined mesh and the solution. This output can be viewed
    #     later using GLVis: "glvis -m refined.mesh -g sol.gf".
    mesh.Print("refined.mesh", 8)
    x.Save("sol.gf", 8)

    # 15. Send the solution by socket to a GLVis server.
    if visualization:
        sol_sock = mfem.socketstream("localhost", 19916)
        sol_sock.precision(8)
        sol_sock << "solution\n" << mesh << x
        sol_sock << "window_title 'Solution'\n"
        sol.sock.flush()

        flux_sock = mfem.socketstream("localhost", 19916)
        flux_sock.precision(8)
        flux_sock << "solution\n" << mesh << flux
        flux_sock << "keys vvv\n"
        flux_sock << "window_geometry 402 0 400 350\n"
        flux_sock << "window_title 'Flux'\n"
        flux_sock.flush()


if __name__ == "__main__":
    from mfem.common.arg_parser import ArgParser

    parser = ArgParser(description='Ex29 (Solving PDEs on embeded surfaces)')
    parser.add_argument('-mt', '--mesh-type',
                        default=4,
                        action='store', type=int,
                        help="Mesh type: 3 - Triangular, 4 - Quadrilateral.")
    parser.add_argument('-mo', '--mesh-order',
                        default=3,
                        action='store', type=int,
                        help="Geometric order of the curved mesh.")
    parser.add_argument("-r", "--refine",
                        action='store', type=int, default=0,
                        help="Number of times to refine the mesh uniformly.")
    parser.add_argument('-o', '--order',
                        action='store', default=3, type=int,
                        help="Finite element order (polynomial degree) or -1 for isoparametric space.")
    parser.add_argument('-sc', '--static-condensation',
                        action='store_true',
                        default=False,
                        help="Enable static condensation.")
    parser.add_argument('-vis', '--visualization',
                        action='store_true',
                        help='Enable GLVis visualization')

    args = parser.parse_args()
    parser.print_options(args)

    run(order=args.order,
        ref_levels=args.refine,
        static_cond=args.static_condensation,
        mesh_type=args.mesh_type,
        mesh_order=args.mesh_order,
        visualization=args.visualization,)



from __future__ import print_function
import os
from os.path import expanduser, join
import sys
import numpy as np
import io
from mfem import path as mfem_path

if len(sys.argv) > 1 and sys.argv[1] == '-p':   
    import mfem.par as mfem
    use_parallel = True
    from mfem.common.mpi_debug import nicePrint as print
    from mpi4py import MPI
    myid  = MPI.COMM_WORLD.rank
    
else:
    import mfem.ser as mfem
    use_parallel = False
    myid = 0

def run_test():

    dir = os.path.dirname(os.path.abspath(__file__))    
    meshfile = os.path.join(dir, "../data/amr-quad.mesh")
    mesh = mfem.Mesh(meshfile, 1, 1)
    dim = mesh.Dimension()
    sdim = mesh.SpaceDimension()
    order = 2
    
    fec = mfem.DG_FECollection(order, dim)
    #fec = mfem.H1_FECollection(order, dim)    
    fes = mfem.FiniteElementSpace(mesh, fec, 1)
    print('Number of finite element unknowns: '+
          str(fes.GetTrueVSize()))
    print('Size of DoFVinite element unknowns: '+
          str(fes.GetNDofs()))

    points = np.zeros((fes.GetTrueVSize(), sdim))

    R = fes.GetConformingRestriction()
    if R is not None:    
        VDof2TDof = np.zeros(fes.GetNDofs(), dtype=int)
        for i, j in enumerate(R.GetJArray()):
            VDof2TDof[j] = i
        TDof2Vdof = R.GetJArray().copy()
    else:
        VDof2TDof = None
        TDof2VDof = None
        
    for j in range(fes.GetNE()):
        el = fes.GetFE(j)
        tr = fes.GetElementTransformation(j)
        vdofs = fes.GetElementVDofs(j)
        
        tdofs= vdofs if VDof2TDof is None else [VDof2TDof[k] for k in vdofs]
        
        ir = el.GetNodes()
        for k, tdof in enumerate(tdofs):
            points[tdof] = tr.Transform(ir.IntPoint(k))

    # search for tdof = 10
    tdof = 10
    vdof = tdof if TDof2VDof is None else TDof2Vdof[tdof]
    fes.BuildDofToArrays()
    j = fes.GetElementForDof(vdof)
    el = fes.GetFE(j)
    tr = fes.GetElementTransformation(j)
    vdofs = fes.GetElementVDofs(j)
    ir = el.GetNodes()
    for k, vv in enumerate(vdofs):
        if vdof == vv:
           pp = tr.Transform(ir.IntPoint(k))
           break
       
    #print(points)
    #print(pp)
if __name__=='__main__':
    run_test()



'''
   MFEM example 2

   See c++ version in the MFEM library for more detail
'''
import os
from os.path import expanduser, join, dirname
import numpy as np

import mfem.ser as mfem


def run(order=1, static_cond=False,
        meshfile='', visualization=False,
        device='cpu', pa=False):
    '''
    run ex2
    '''
    device = mfem.Device(device)
    device.Print()

    mesh = mfem.Mesh(meshfile, 1, 1)
    dim = mesh.Dimension()

    #   3. Refine the mesh to increase the resolution. In this example we do
    #      'ref_levels' of uniform refinement. We choose 'ref_levels' to be the
    #      largest number that gives a final mesh with no more than 50,000
    #      elements.

    ref_levels = int(np.floor(np.log(5000. / mesh.GetNE()) / np.log(2.) / dim))
    for x in range(ref_levels):
        mesh.UniformRefinement()

    # 5. Define a finite element space on the mesh. Here we use vector finite
    #   elements, i.e. dim copies of a scalar finite element space. The vector
    #   dimension is specified by the last argument of the FiniteElementSpace
    #   constructor. For NURBS meshes, we use the (degree elevated) NURBS space
    #   associated with the mesh nodes.
    fec = mfem.H1_FECollection(order, dim)
    fespace = mfem.FiniteElementSpace(mesh, fec, dim)

    print("Number of finite element unknowns: " + str(fespace.GetTrueVSize()))

    # 6. Determine the list of true (i.e. conforming) essential boundary dofs.
    #    In this example, the boundary conditions are defined by marking only
    #    boundary attribute 1 from the mesh as essential and converting it to a
    #    list of true dofs.
    ess_tdof_list = mfem.intArray()
    ess_bdr = mfem.intArray([1] + [0] * (mesh.bdr_attributes.Max() - 1))
    fespace.GetEssentialTrueDofs(ess_bdr, ess_tdof_list)
    # 7. Set up the linear form b(.) which corresponds to the right-hand side of
    #    the FEM linear system. In this case, b_i equals the boundary integral
    #    of f*phi_i where f represents a "pull down" force on the Neumann part
    #    of the boundary and phi_i are the basis functions in the finite element
    #    fespace. The force is defined by the VectorArrayCoefficient object f,
    #    which is a vector of Coefficient objects. The fact that f is non-zero
    #    on boundary attribute 2 is indicated by the use of piece-wise constants
    #    coefficient for its last component.
    f = mfem.VectorArrayCoefficient(dim)
    for i in range(dim - 1):
        f.Set(i, mfem.ConstantCoefficient(0.0))

    pull_force = mfem.Vector([0] * mesh.bdr_attributes.Max())
    pull_force[1] = -1.0e-2
    f.Set(dim - 1, mfem.PWConstCoefficient(pull_force))

    b = mfem.LinearForm(fespace)
    b.AddBoundaryIntegrator(mfem.VectorBoundaryLFIntegrator(f))
    print('r.h.s...')
    b.Assemble()
    # 8. Define the solution vector x as a finite element grid function
    #    corresponding to fespace. Initialize x with initial guess of zero,
    #    which satisfies the boundary conditions.
    x = mfem.GridFunction(fespace)
    x.Assign(0.0)
    # 9. Set up the bilinear form a(.,.) on the finite element space
    #    corresponding to the linear elasticity integrator with piece-wise
    #    constants coefficient lambda and mu.
    #    (Here lambda is renamed by lamb since it is reserved in python)
    lamb = mfem.Vector(mesh.attributes.Max())
    lamb.Assign(1.0)
    lamb[0] = lamb[1] * 50
    lambda_func = mfem.PWConstCoefficient(lamb)
    mu = mfem.Vector(mesh.attributes.Max())
    mu.Assign(1.0)
    mu[0] = mu[1] * 50
    mu_func = mfem.PWConstCoefficient(mu)
    a = mfem.BilinearForm(fespace)
    a.AddDomainIntegrator(mfem.ElasticityIntegrator(lambda_func, mu_func))
    # 10. Assemble the bilinear form and the corresponding linear system,
    #     applying any necessary transformations such as: eliminating boundary
    #     conditions, applying conforming constraints for non-conforming AMR,
    #     static condensation, etc.
    print('matrix...')
    if (static_cond):
        a.EnableStaticCondensation()
    a.Assemble()

    A = mfem.OperatorPtr()
    B = mfem.Vector()
    X = mfem.Vector()
    a.FormLinearSystem(ess_tdof_list, x, b, A, X, B)
    print('...done')  # Here, original version calls hegith, which is not
    # defined in the header...!?
    print("Size of linear system: " + str(A.Height()))

    # 10. Solve
    AA = mfem.OperatorHandle2SparseMatrix(A)
    M = mfem.GSSmoother(AA)
    mfem.PCG(AA, M, B, X, 1, 500, 1e-8, 0.0)

    # 11. Recover the solution as a finite element grid function.
    a.RecoverFEMSolution(X, b, x)

    # 13. For non-NURBS meshes, make the mesh curved based on the finite element
    #     space. This means that we define the mesh elements through a fespace
    #     based transformation of the reference element. This allows us to save
    #     the displaced mesh as a curved mesh when using high-order finite
    #     element displacement field. We assume that the initial mesh (read from
    #     the file) is not higher order curved mesh compared to the chosen FE
    #     space.
    if not mesh.NURBSext:
        mesh.SetNodalFESpace(fespace)

    # 14. Save the displaced mesh and the inverted solution (which gives the
    #     backward displacements to the original grid). This output can be
    #     viewed later using GLVis: "glvis -m displaced.mesh -g sol.gf".
    nodes = mesh.GetNodes()
    nodes += x
    x *= -1
    mesh.Print('displaced.mesh', 8)
    x.Save('sol.gf', 8)

    if (visualization):
        sol_sock = mfem.socketstream("localhost", 19916)
        sol_sock.precision(8)
        sol_sock.send_solution(mesh, x)


if __name__ == "__main__":
    from mfem.common.arg_parser import ArgParser
    parser = ArgParser(description='Ex2 (Linear elasticity)')
    parser.add_argument('-m', '--mesh',
                        default='beam-tri.mesh',
                        action='store', type=str,
                        help='Mesh file to use.')
    parser.add_argument('-vis', '--visualization',
                        action='store_true',
                        help='Enable GLVis visualization')
    parser.add_argument('-o', '--order',
                        action='store', default=1, type=int,
                        help="Finite element order (polynomial degree) or -1 for isoparametric space.")
    parser.add_argument('-sc', '--static-condensation',
                        action='store_true',
                        help="Enable static condensation.")
    parser.add_argument("-pa", "--partial-assembly",
                        action='store_true',
                        help="Enable Partial Assembly.")
    parser.add_argument("-d", "--device",
                        default="cpu", type=str,
                        help="Device configuration string, see Device::Configure().")

    args = parser.parse_args()
    parser.print_options(args)

    order = args.order
    static_cond = args.static_condensation

    meshfile = expanduser(
        join(dirname(__file__), '..', 'data', args.mesh))

    visualization = args.visualization
    device = args.device
    pa = args.partial_assembly

    run(order=order,
        static_cond=static_cond,
        meshfile=meshfile,
        visualization=visualization,
        device=device,
        pa=pa)



from __future__ import print_function

import os
import sys
import time
import subprocess as sp

modules = ["point", "segment", "mesh"]


def run_file(command):
    print('command',command)
    t1 = time.time()    
    p = sp.Popen(command,stdout=sp.PIPE, stderr=sp.STDOUT, stdin=sp.PIPE)
    lines = p.stdout.readlines()
    t2 = time.time()
    sys.stdout.flush()
    print(lines)
    return t2-t1, lines

def run_test(serial = True, np=2):
    file = ["test_"+x+".py" for x in modules]
    
    if not serial:
        comm_mpi = ["mpirun", "-np", str(np)]
    else:
        comm_mpi = []

    for f in file:
        comm = comm_mpi + [sys.executable,  f]
        if not serial:
            comm = comm.append("-p")


        run_file(comm)

if __name__=='__main__':
    if len(sys.argv) > 1 and sys.argv[1] == '-p':   
        run_test(serial = False)
    else:
        run_test(serial = True)




'''
   MFEM example 30

   See c++ version in the MFEM library for more detail 
'''
import mfem.par as mfem
from mfem.par import intArray
import os
from os.path import expanduser, join
import numpy as np
from numpy import sin, array

try:
    from numba import jit
except:
    assert False, "This example requires numba"

from mpi4py import MPI
num_procs = MPI.COMM_WORLD.size
myid = MPI.COMM_WORLD.rank
smyid = '.'+'{:0>6d}'.format(myid)


@mfem.jit.scalar()
def affine_coeff(p):
    if p[0] < 0.0:
        return 1 + p[0] + p[1]
    else:
        return 1


@mfem.jit.scalar()
def jump_coeff(p):
    norm = np.sqrt(p[0]**2 + p[1]**2)
    if norm > 0.4 and norm < 0.6:
        return 1.0
    else:
        return 5.0


@mfem.jit.scalar()
def singular_coeff(p):
    x = p[0]
    y = p[1]
    alpha = 1000.0
    xc = 0.75
    yc = 0.5
    r0 = 0.7
    r = np.sqrt((x - xc)**2.0 + (y - yc)**2)
    num = - (alpha - alpha**3 * (r**2 - r0**2))
    denom = (r * (alpha**2 * r0**2 + alpha**2 * r**2 -
                  2 * alpha**2 * r0 * r + 1.0))**2
    denom = max([denom, 1e-8])
    return num / denom


def run(nc_limit=1,
        order=1,
        meshfile='',
        visualization=True,
        max_elems=100000,
        enriched_order=5,
        osc_threshold=1e-3,
        nc_simplices=True):

    device = mfem.Device('cpu')
    if myid == 0:
        device.Print()

    mesh = mfem.Mesh(meshfile, 1, 1)

    if mesh.NURBSext:
        for i in range(2):
            mesh.UniformRefinement()
        mesh.SetCurvature(2)

    mesh.EnsureNCMesh(nc_simplices)

    pmesh = mfem.ParMesh(MPI.COMM_WORLD, mesh)
    mesh.Clear()

    coeffrefiner = mfem.CoefficientRefiner(affine_coeff, order)

    if (visualization):
        sol_sock = mfem.socketstream("localhost", 19916)

    irs = [mfem.IntRules.Get(i, 2*order + enriched_order)
           for i in range(mfem.Geometry.NumGeom)]

    coeffrefiner.SetIntRule(irs)
    coeffrefiner.SetMaxElements(max_elems)
    coeffrefiner.SetThreshold(osc_threshold)
    coeffrefiner.SetNCLimit(nc_limit)
    coeffrefiner.PrintWarnings()

    coeffrefiner.PreprocessMesh(pmesh)

    globalNE = pmesh.GetGlobalNE()
    osc = coeffrefiner.GetOsc()
    if myid == 0:
        print()
        print("Function 0 (affine)")
        print("Number of Elements " + str(globalNE))
        print("Osc error " + "{:g}".format(osc))

    # 8. Preprocess mesh to control osc (jump function).
    coeffrefiner.ResetCoefficient(jump_coeff)
    coeffrefiner.PreprocessMesh(pmesh)

    globalNE = pmesh.GetGlobalNE()
    osc = coeffrefiner.GetOsc()
    if myid == 0:
        print("")
        print("Function 1 (discontinuous)")
        print("Number of Elements " + str(globalNE))
        print("Osc error " + "{:g}".format(osc))

    # 9. Preprocess mesh to control osc (singular function).
    coeffrefiner.ResetCoefficient(singular_coeff)
    coeffrefiner.PreprocessMesh(pmesh)

    globalNE = pmesh.GetGlobalNE()
    osc = coeffrefiner.GetOsc()
    if myid == 0:
        print("")
        print("Function 2 (singular)")
        print("Number of Elements " + str(globalNE))
        print("Osc error " + "{:g}".format(osc))

    if visualization:
        sol_sock.precision(8)
        sol_sock << "parallel " << num_procs << " " << myid << "\n"
        sol_sock << "mesh\n" << pmesh
        sol_sock.flush()


if __name__ == "__main__":
    from mfem.common.arg_parser import ArgParser

    parser = ArgParser(description='Ex30p (Parallel coefficient refinere)')
    parser.add_argument('-m', '--mesh',
                        default="star.mesh",
                        action='store', type=str,
                        help='Mesh file to use.')
    parser.add_argument('-o', '--order',
                        action='store', default=1, type=int,
                        help="Finite element order (polynomial degree)")
    parser.add_argument("-l", "--nc_limit",
                        action='store',
                        type=int,
                        default=1,
                        help="Maximum level of hanging nodes.")
    parser.add_argument('-me', '--max-elems',
                        action='store',
                        type=int, default=100000,
                        help="Stop after reaching this many elements",)
    parser.add_argument('-vis', '--visualization',
                        action='store_true',
                        help='Enable GLVis visualization')
    parser.add_argument("-e", "--error",
                        action='store',
                        default=1e-3,
                        type=float,
                        help="relative data oscillation threshold.")
    parser.add_argument("-eo", "--enriched-order",
                        type=int,
                        default=5,
                        help="Enriched quadrature order.")
    parser.add_argument("-ns", "--nonconforming-simplices",
                        action='store_true',
                        default=True,
                        help="For simplicial meshes, enable nonconforming refinement")
    parser.add_argument("-cs", "--conforming-simplices",
                        action='store_true',
                        default=False,
                        help="For simplicial meshes, disable nonconforming refinement")

    args = parser.parse_args()
    parser.print_options(args)

    nc_simplices = args.nonconforming_simplices
    if args.conforming_simplices:
        nc_simplices = False

    meshfile = expanduser(
        join(os.path.dirname(__file__), '..', 'data', args.mesh))
    print(meshfile)

    run(nc_limit=args.nc_limit,
        order=args.order,
        meshfile=meshfile,
        visualization=args.visualization,
        max_elems=args.max_elems,
        enriched_order=args.enriched_order,
        osc_threshold=args.error,
        nc_simplices=nc_simplices)



from __future__ import print_function

import io
import gzip

import numpy as np

import os
from os.path import expanduser, join
import sys
import numpy as np
import io

if len(sys.argv) > 1 and sys.argv[1] == '-p':   
    import mfem.par as mfem
    use_parallel = True
    from mfem.common.mpi_debug import nicePrint as print
    from mpi4py import MPI
    myid  = MPI.COMM_WORLD.rank
    
else:
    import mfem.ser as mfem
    use_parallel = False
    myid = 0

def check(a, b, msg):
    assert len(a) == len(b), msg
    assert np.sum(np.abs(a-b)) == 0, msg

def check_mesh(m1, m2, msg):
    assert m1.GetNE() == m2.GetNE(), msg
    assert m1.GetNBE() == m2.GetNBE(), msg
    assert m1.GetNV() == m2.GetNV(), msg

def run_test():    
    mesh = mfem.Mesh(10, 10, "TRIANGLE")

    fec = mfem.H1_FECollection(1, mesh.Dimension())
    fespace = mfem.FiniteElementSpace(mesh, fec)

    gf = mfem.GridFunction(fespace)
    gf.Assign(1.0)
    odata = gf.GetDataArray().copy()

    visit_dc = mfem.VisItDataCollection("test_gf", mesh)
    visit_dc.RegisterField("gf", gf)
    visit_dc.Save()

    mesh2 = mfem.Mesh("test_gf_000000/mesh.000000")
    check_mesh(mesh, mesh2, ".mesh2 does not agree with original")

    gf2 = mfem.GridFunction(mesh2, "test_gf_000000/gf.000000")
    odata2 = gf2.GetDataArray().copy()
    check(odata, odata2, "odata2 file does not agree with original")

    visit_dc = mfem.VisItDataCollection("test_gf_gz", mesh)
    visit_dc.SetCompression(True)
    visit_dc.RegisterField("gf", gf)
    visit_dc.Save()

    with gzip.open("test_gf_gz_000000/mesh.000000", 'rt') as f:
       sio = io.StringIO(f.read())
    mesh3 = mfem.Mesh(sio)
    check_mesh(mesh, mesh3, ".mesh3 does not agree with original")

    with gzip.open("test_gf_gz_000000/gf.000000", 'rt') as f:
       sio = io.StringIO(f.read())

    # This is where the error is:
    gf3 = mfem.GridFunction(mesh3, sio)
    odata3 = gf3.GetDataArray()
    check(odata, odata3, "gf3 file does not agree with original")

    mesh4 = mfem.Mesh("test_gf_gz_000000/mesh.000000")
    check_mesh(mesh, mesh4, ".mesh4 does not agree with original")

    gf4 = mfem.GridFunction(mesh3, "test_gf_gz_000000/gf.000000")
    odata4 = gf4.GetDataArray()

    check(odata, odata4, "gf3 file does not agree with original")

    print("PASSED")    

if __name__=='__main__':
    run_test()



'''
   MFEM example 7

   See c++ version in the MFEM library for more detail 
'''
from mfem import path
import mfem.par as mfem
from mfem.par import intArray
from mpi4py import MPI
from os.path import expanduser, join
import numpy as np
from numpy import sin, cos, exp


num_proc = MPI.COMM_WORLD.size
myid = MPI.COMM_WORLD.rank

elem_type = 1
ref_levels = 2
par_ref_levels = 2
amr = 0
order = 2
always_snap = False

device = mfem.Device('cpu')
if myid == 0:
    device.Print()

if elem_type == 1:
    Nvert = 8
    Nelem = 6
else:
    Nvert = 6
    Nelem = 8


mesh = mfem.Mesh(2, Nvert, Nelem, 0, 3)

if elem_type == 0:
    tri_v = [[1.,  0.,  0.], [0.,  1.,  0.], [-1.,  0.,  0.],
             [0., -1.,  0.], [0.,  0.,  1.], [0.,  0., -1.]]
    tri_e = [[0, 1, 4], [1, 2, 4], [2, 3, 4], [3, 0, 4],
             [1, 0, 5], [2, 1, 5], [3, 2, 5], [0, 3, 5]]
    for j in range(Nvert):
        mesh.AddVertex(tri_v[j])
    for j in range(Nelem):
        mesh.AddTriangle(tri_e[j], j+1)
    mesh.FinalizeTriMesh(1, 1, True)
else:
    quad_v = [[-1, -1, -1], [+1, -1, -1], [+1, +1, -1], [-1, +1, -1],
              [-1, -1, +1], [+1, -1, +1], [+1, +1, +1], [-1, +1, +1]]

    quad_e = [[3, 2, 1, 0], [0, 1, 5, 4], [1, 2, 6, 5],
              [2, 3, 7, 6], [3, 0, 4, 7], [4, 5, 6, 7]]
    for j in range(Nvert):
        mesh.AddVertex(quad_v[j])
    for j in range(Nelem):
        mesh.AddQuad(quad_e[j], j+1)
    mesh.FinalizeQuadMesh(1, 1, True)

#  Set the space for the high-order mesh nodes.
fec = mfem.H1_FECollection(order, mesh.Dimension())
nodal_fes = mfem.FiniteElementSpace(mesh, fec, mesh.SpaceDimension())
mesh.SetNodalFESpace(nodal_fes)


def SnapNodes(mesh):
    nodes = mesh.GetNodes()
    node = mfem.Vector(mesh.SpaceDimension())

    for i in np.arange(nodes.FESpace().GetNDofs()):
        for d in np.arange(mesh.SpaceDimension()):
            node[d] = nodes[nodes.FESpace().DofToVDof(i, d)]

        node /= node.Norml2()
        for d in range(mesh.SpaceDimension()):
            nodes[nodes.FESpace().DofToVDof(i, d)] = node[d]


#  3. Refine the mesh while snapping nodes to the sphere.
for l in range(ref_levels+1):
    if l > 0:
        mesh.UniformRefinement()
    if always_snap:
        SnapNodes(mesh)
if amr == 1:
    for l in range(5):
        mesh.RefineAtVertex(mfem.Vertex(0, 0, 1))
    SnapNodes(mesh)
elif amr == 2:
    for l in range(4):
        mesh.RandomRefinement(0.5)
    SnapNodes(mesh)

pmesh = mfem.ParMesh(MPI.COMM_WORLD, mesh)
for i in range(par_ref_levels):
    pmesh.UniformRefinement()
    if always_snap:
        SnapNodes(pmesh)
if (not always_snap) or (par_ref_levels < 1):
    SnapNodes(pmesh)
if amr == 1:
    for l in range(5):
        pmesh.RefineAtVertex(mfem.Vertex(0, 0, 1))
    SnapNodes(pmesh)
elif amr == 2:
    for l in range(4):
        pmesh.RandomRefinement(0.5)
    SnapNodes(pmesh)


#  4. Define a finite element space on the mesh. Here we use isoparametric
#      finite elements -- the same as the mesh nodes.
fespace = mfem.ParFiniteElementSpace(pmesh, fec)
size = fespace.GlobalTrueVSize()
if myid == 0:
    print("Number of unknowns: " + str(size))

#   5. Set up the linear form b(.) which corresponds to the right-hand side of
#      the FEM linear system, which in this case is (1,phi_i) where phi_i are
#      the basis functions in the finite element fespace.


class analytic_rhs(mfem.PyCoefficient):
    def EvalValue(self, x):
        l2 = np.sum(x**2)
        return 7.*x[0]*x[1]/l2


class analytic_solution(mfem.PyCoefficient):
    def EvalValue(self, x):
        l2 = np.sum(x**2)
        return x[0]*x[1]/l2


b = mfem.ParLinearForm(fespace)
one = mfem.ConstantCoefficient(1.0)
rhs_coef = analytic_rhs()
sol_coef = analytic_solution()
b.AddDomainIntegrator(mfem.DomainLFIntegrator(rhs_coef))
b.Assemble()

#  6. Define the solution vector x as a finite element grid function
#     corresponding to fespace. Initialize x with initial guess of zero.
x = mfem.ParGridFunction(fespace)
x.Assign(0.0)

#  7. Set up the bilinear form a(.,.) on the finite element space
#     corresponding to the Laplacian operator -Delta, by adding the Diffusion
#     and Mass domain integrators.
a = mfem.ParBilinearForm(fespace)
a.AddDomainIntegrator(mfem.DiffusionIntegrator(one))
a.AddDomainIntegrator(mfem.MassIntegrator(one))

#  8. Assemble the linear system, apply conforming constraints, etc.
a.Assemble()
A = mfem.HypreParMatrix()
B = mfem.Vector()
X = mfem.Vector()
empty_tdof_list = mfem.intArray()

a.FormLinearSystem(empty_tdof_list, x, b, A, X, B)

amg = mfem.HypreBoomerAMG(A)
pcg = mfem.HyprePCG(A)
pcg.SetTol(1e-12)
pcg.SetMaxIter(200)
pcg.SetPrintLevel(2)
pcg.SetPreconditioner(amg)
pcg.Mult(B, X)

# 10. Recover the solution as a finite element grid function.
a.RecoverFEMSolution(X, b, x)

# 12. Compute and print the L^2 norm of the error.
err = x.ComputeL2Error(sol_coef)
if myid == 0:
    print("L2 norm of error = " + str(err))


smyid = '{:0>6d}'.format(myid)
mesh_name = "sphere_refined."+smyid
sol_name = "sol."+smyid

pmesh.Print(mesh_name, 8)
x.Save(sol_name, 8)



'''
 subroutines for numba_coefficeint
'''


def generate_caller_scalar(setting):
    '''
    generate a callder function on the fly

    ex)
    if setting is
        {"isdepcomplex": (True, False), "kinds": (1, 0),
                       "output": True, size: (10, 1)}

    def _caller(ptx, sdim, data):
        ptx = farray(ptx, (sdim,), np.float64)      # for position
        arr0r = farray(data[0], (10,), np.float64)
        arr0i = farray(data[1], (10,), np.float64)
        arr0 = arr0r+1j*arr0i

        arr1 = farray(data[0], (1,), np.float64)

        params = (arr0, arr1)
        return (inner_func(ptx, *params))

    here inner_func is a function user provided.

    '''
    if setting['td']:
        text = ['def _caller(ptx, sdim, t, data):']
    else:
        text = ['def _caller(ptx, sdim, data):']

    text.append("    ptx = farray(ptx, (sdim,), np.float64)")
    count = 0
    params_line = '    params = ('

    for s, kind, size in zip(setting['isdepcomplex'], setting['kinds'], setting["sizes"]):
        if not isinstance(size, tuple):
            size = (size, )

        if s:
            t1 = '    arrr' + \
                str(count) + ' = farray(data[' + \
                str(count) + "], "+str(size) + ", np.float64)"
            t2 = '    arri' + \
                str(count) + ' = farray(data[' + \
                str(count+1) + "], "+str(size) + ", np.float64)"
            t3 = '    arr'+str(count) + ' = arrr' + \
                str(count) + "+1j*arri" + str(count)

            #if len(size) == 1 and size[0] == 1:
            if kind == 0:
                t1 += '[0]'
                t2 += '[0]'

            text.extend((t1, t2, t3))
            params_line += 'arr'+str(count)+','
            count = count + 2
        else:
            t = '    arr' + \
                str(count) + ' = farray(data[' + \
                str(count) + "], "+str(size) + ", np.float64)"

            #if len(size) == 1 and size[0] == 1:
            if kind == 0:            
                t += '[0]'

            text.append(t)

            params_line += 'arr'+str(count)+','
            count = count + 1

    params_line += ')'

    text.append(params_line)
    if setting["td"]:
        text.append("    return (inner_func(ptx, t, *params))")
    else:
        text.append("    return (inner_func(ptx, *params))")
    return '\n'.join(text)


def generate_caller_array_oldstyle(setting):
    '''
    generate a callder function on the fly
    ex)
    if setting is
        {"isdepcomplex": (True, False), "kinds": (1, 0),
                       "output": True, size: ((3, 3), 1), outsize: (2, 2) }
    def _caller(ptx, sdim, data, out_):
        ptx = farray(ptx, (sdim,), np.float64)      # for position
        arr0r = farray(data[0], (3, 3), np.float64)
        arr0i = farray(data[1], (3, 3), np.float64)
        arr0 = arr0r+1j*arr0i
        arr1 = farray(data[0], (1,), np.float64)
        out = farray(out_, (2, 2), np.complex128)
        params = (arr0, arr1, out,)
        return (inner_func(ptx, *params))
    here inner_func is a function user provided.
    '''
    if setting['td']:
        text = ['def _caller(ptx, sdim, t, data, out_):']
    else:
        text = ['def _caller(ptx, sdim, data, out_):']
    text.append("    ptx = farray(ptx, (sdim,), np.float64)")
    count = 0
    params_line = '    params = ('

    for s, kind, size in zip(setting['isdepcomplex'], setting['kinds'], setting["sizes"]):
        if not isinstance(size, tuple):
            size = (size, )

        if s:
            t1 = '    arrr' + \
                str(count) + ' = farray(data[' + \
                str(count) + "], "+str(size) + ", np.float64)"
            t2 = '    arri' + \
                str(count) + ' = farray(data[' + \
                str(count+1) + "], "+str(size) + ", np.float64)"

            #if len(size) == 1 and size[0] == 1:
            if kind == 0:                
                t1 += '[0]'
                t2 += '[0]'

            t3 = '    arr'+str(count) + ' = arrr' + \
                str(count) + "+1j*arri" + str(count)

            text.extend((t1, t2, t3))
            params_line += 'arr'+str(count)+','
            count = count + 2
        else:
            t = '    arr' + \
                str(count) + ' = farray(data[' + \
                str(count) + "], "+str(size) + ", np.float64)"

            #if len(size) == 1 and size[0] == 1:
            if kind == 0:                                        
                t += '[0]'

            text.append(t)

            params_line += 'arr'+str(count)+','
            count = count + 1

    outsize = setting["outsize"]
    if setting["output"]:
        t = '    out = farray(out_,' + str(outsize) + ", np.complex128)"
    else:
        t = '    out = farray(out_,' + str(outsize) + ", np.float64)"
    text.append(t)

    params_line += 'out, )'

    text.append(params_line)
    if setting["td"]:
        text.append("    return (inner_func(ptx, t, *params))")
    else:
        text.append("    return (inner_func(ptx, *params))")
    return '\n'.join(text)


def generate_caller_array(setting):
    '''
    generate a following callder function from setting

    ex)
    if setting is
        {"isdepcomplex": (True, False), "kinds": (1, 0),
                       "output": True, size: ((3, 3), 1), outsize: (2, 2) }

    def _caller(ptx, sdim, data, out_):
        ptx = farray(ptx, (sdim,), np.float64)      # for position
        arr0r = farray(data[0], (3, 3), np.float64)
        arr0i = farray(data[1], (3, 3), np.float64)
        arr0 = arr0r+1j*arr0i

        arr1 = farray(data[0], (1,), np.float64)

        out = farray(out_, (2, 2), np.complex128)

        params = (arr0, arr1, )

        ret = inner_func(ptx, *params)
        for i0 in range(2):
           for i1 in range(2):
              ret[i0,i1] = out[i0, i1]

    here inner_func is a function user provided.

    '''
    if setting['td']:
        text = ['def _caller(ptx, sdim, t, data, out_):']
    else:
        text = ['def _caller(ptx, sdim, data, out_):']
    text.append("    ptx = farray(ptx, (sdim,), np.float64)")
    count = 0
    params_line = '    params = ('

    for s, kind, size in zip(setting['isdepcomplex'], setting['kinds'], setting["sizes"]):
        if not isinstance(size, tuple):
            size = (size, )

        if s:
            t1 = '    arrr' + \
                str(count) + ' = farray(data[' + \
                str(count) + "], "+str(size) + ", np.float64)"
            t2 = '    arri' + \
                str(count) + ' = farray(data[' + \
                str(count+1) + "], "+str(size) + ", np.float64)"

            #if len(size) == 1 and size[0] == 1:
            if kind == 0:                                
                t1 += '[0]'
                t2 += '[0]'

            t3 = '    arr'+str(count) + ' = arrr' + \
                str(count) + "+1j*arri" + str(count)

            text.extend((t1, t2, t3))
            params_line += 'arr'+str(count)+','
            count = count + 2
        else:
            t = '    arr' + \
                str(count) + ' = farray(data[' + \
                str(count) + "], "+str(size) + ", np.float64)"

            #if len(size) == 1 and size[0] == 1:
            if kind == 0:                                                
                t += '[0]'

            text.append(t)

            params_line += 'arr'+str(count)+','
            count = count + 1

    outsize = setting["outsize"]
    if setting["output"]:
        t = '    out = farray(out_,' + str(outsize) + ", np.complex128)"
    else:
        t = '    out = farray(out_,' + str(outsize) + ", np.float64)"
    text.append(t)
    '''
    params_line += 'out, )'
    '''
    params_line += ')'
    text.append(params_line)
    if setting["td"]:
        text.append("    ret = inner_func(ptx, t, *params)")
    else:
        text.append("    ret = inner_func(ptx, *params)")

    idx_text = ""
    for k, s in enumerate(setting["outsize"]):
        text.append("    " + " "*k + "for i" + str(k) +
                    " in range(" + str(s) + "):")
        idx_text = idx_text + "i"+str(k)+","
    text.append("     " + " "*len(setting["outsize"]) +
                "out["+idx_text + "]=ret[" + idx_text + "]")

    return '\n'.join(text)


def generate_signature_scalar(setting):
    '''
    generate a signature to numba-compile a user scalar function

    ex)
    when user function is
        func(ptx, complex_array, float_scalar)

    setting is
        {"isdepcomplex": (2, 1), "kinds": (1, 0), "output": 2}

    output is
         types.complex128(types.double[:], types.complex128[:], types.double,)

    user function is

    '''
    sig = ''
    if setting['output']:
        sig += 'types.complex128(types.double[:], '
    else:
        sig += 'types.float64(types.double[:], '

    if setting['td']:
        sig += 'types.double, '

    for s, kind, in zip(setting['isdepcomplex'], setting['kinds'],):
        if s:
            if kind == 0:
                sig += 'types.complex128,'
            elif kind == 1:
                sig += 'types.complex128[:], '
            else:
                sig += 'types.complex128[:, :], '
        else:
            if kind == 0:
                sig += 'types.double,'
            elif kind == 1:
                sig += 'types.double[:], '
            else:
                sig += 'types.double[:, :], '

    sig = sig + ")"
    return sig


def generate_signature_array_oldstyle(setting):
    '''
    generate a signature to numba-compile a user scalar function

    ex)
    when user function is
        func(ptx, complex_array, float_scalar, complex_output_array_)

    setting is
        {"isdepcomplex": (2, 1), "kinds": (1, 0), "output": 2, "outkind": 2}

    output is
         types.void(types.double[:], types.complex128[::],
                    types.double, types.complex128[:, :])

    user function is

    '''
    sig = 'types.void(types.double[:], '
    if setting['td']:
        sig += 'types.double, '

    for s, kind, in zip(setting['isdepcomplex'], setting['kinds'],):
        if s:
            if kind == 0:
                sig += 'types.complex128,'
            elif kind == 1:
                sig += 'types.complex128[:], '
            else:
                sig += 'types.complex128[:,:], '
        else:
            if kind == 0:
                sig += 'types.double,'
            elif kind == 1:
                sig += 'types.double[:], '
            else:
                sig += 'types.double[:, :], '

    if setting['output']:
        if setting['outkind'] == 1:
            sig += 'types.complex128[:], '
        else:
            sig += 'types.complex128[:, :], '
    else:
        if setting['outkind'] == 1:
            sig += 'types.double[:], '
        else:
            sig += 'types.double[:, :], '
    sig = sig + ")"
    return sig


def generate_signature_array(setting):
    '''
    generate a signature to numba-compile a user scalar function

    ex)
    when user function is
        func(ptx, complex_array, float_scalar)

    setting is
        {"isdepcomplex": (2, 1), "kinds": (1, 0), "output": 2}

    output is
         types.complex128[:, :](types.double[:], types.complex128[:], types.double,)

    user function is

    '''
    sig = ''
    if setting['output']:
        if setting['outkind'] == 1:
            sig += 'types.complex128[:](types.double[:], '
        else:
            sig += 'types.complex128[:,:](types.double[:], '
    else:
        if setting['outkind'] == 1:
            sig += 'types.float64[:](types.double[:], '
        else:
            sig += 'types.float64[:,:](types.double[:], '
    if setting['td']:
        sig += 'types.double, '

    for s, kind, in zip(setting['isdepcomplex'], setting['kinds'],):
        if s:
            if kind == 0:
                sig += 'types.complex128,'
            elif kind == 1:
                sig += 'types.complex128[:], '
            else:
                sig += 'types.complex128[:, :], '
        else:
            if kind == 0:
                sig += 'types.double,'
            elif kind == 1:
                sig += 'types.double[:], '
            else:
                sig += 'types.double[:, :], '

    sig = sig + ")"
    return sig


def _process_dependencies(dependencies, setting):
    from mfem import mfem_mode
    if mfem_mode == 'serial':
        from mfem.ser import (Coefficient,
                              VectorCoefficient,
                              MatrixCoefficient,
                              IsNumbaCoefficient)
    else:
        from mfem.par import (Coefficient,
                              VectorCoefficient,
                              MatrixCoefficient,
                              IsNumbaCoefficient)

    iscomplex = []
    sizes = []
    kinds = []
    s_coeffs = []
    v_coeffs = []
    m_coeffs = []
    ns_coeffs = []
    nv_coeffs = []
    nm_coeffs = []
    for x in dependencies:
        if isinstance(x, tuple) and x[1] is not None:
            iscomplex.append(1)
            xx = x[0]
            if x[0] is None or x[1] is None:
                assert False, "dependency has to have both real imaginary parts defined"
            if isinstance(xx, VectorCoefficient):
                assert x[0].GetVDim() == x[1].GetVDim(
                ), "real and imaginary has to have the same vdim"
            if isinstance(xx, MatrixCoefficient):
                h1, w1 = x[0].GetHeight(), x[0].GetWidth()
                h2, w2 = x[1].GetHeight(), x[1].GetWidth()
                assert h1 == w1, "matrix must be square"
                assert h1 == h2, "real and imaginary has to have the same vdim"
                assert w1 == w2, "real and imaginary has to have the same vdim"
        else:
            if isinstance(x, tuple):
                # treat complex without imaginary part as real
                x = x[0]
            xx = x
            if IsNumbaCoefficient(xx):
                if xx.IsOutComplex():
                    iscomplex.append(2)
                else:
                    iscomplex.append(0)
            else:
                iscomplex.append(0)

            if isinstance(xx, MatrixCoefficient):
                assert xx.GetHeight() == xx.GetWidth(), "matrix must be square"

        if isinstance(xx, Coefficient):
            kinds.append(0)
            sizes.append(1)
            s_coeffs.append(xx)
            if iscomplex[-1] > 0:
                if IsNumbaCoefficient(xx) and not isinstance(x, tuple):
                    ns_coeffs.append(xx)
                else:
                    s_coeffs.append(x[1])

        elif isinstance(xx, VectorCoefficient):
            kinds.append(1)
            sizes.append(xx.GetVDim())
            v_coeffs.append(xx)
            if iscomplex[-1] > 0:
                if IsNumbaCoefficient(xx) and not isinstance(x, tuple):
                    nv_coeffs.append(xx)
                else:
                    v_coeffs.append(x[1])

        elif isinstance(xx, MatrixCoefficient):
            kinds.append(2)
            sizes.append((xx.GetHeight(), xx.GetWidth()))
            m_coeffs.append(xx)
            if iscomplex[-1] > 0:
                if IsNumbaCoefficient(xx) and not isinstance(x, tuple):
                    nm_coeffs.append(xx)
                else:
                    m_coeffs.append(x[1])

        else:
            assert False, "unknown coefficient type" + str(type(xx))

    setting['s_coeffs'] = s_coeffs
    setting['v_coeffs'] = v_coeffs
    setting['m_coeffs'] = m_coeffs
    setting['ns_coeffs'] = ns_coeffs
    setting['nv_coeffs'] = nv_coeffs
    setting['nm_coeffs'] = nm_coeffs

    return iscomplex, sizes, kinds


def get_setting(outsize, iscomplex=False, dependencies=None, td=False):
    setting = {}
    if iscomplex:
        setting['output'] = True
    else:
        setting['output'] = False

    isdepcomplex, sizes, kinds = _process_dependencies(dependencies, setting)

    setting['isdepcomplex'] = isdepcomplex
    setting['kinds'] = kinds
    setting['outsize'] = outsize
    if isinstance(outsize, tuple):
        setting['outkind'] = len(outsize)
    else:
        setting['outkind'] = 0
    setting['sizes'] = sizes
    setting['td'] = td

    return setting



'''
   MFEM example 31

   See c++ version in the MFEM library for more detail

   Sample runs: mpirun -np 4 python ex31p.py -m ../data/hexagon.mesh -o 2
                mpirun -np 4 python ex31p.py -m ../data/star.mesh 
                mpirun -np 4 python ex31p.py -m ../data/square-disc.mesh -o 2
                mpirun -np 4 python ex31p.py -m ../data/fichera.mesh -o 3 -rs 1 -rp 0
                mpirun -np 4 python ex31p.py -m ../data/square-disc-nurbs.mesh -o 3
                mpirun -np 4 python ex31p.py -m ../data/amr-quad.mesh -o 2 -rs 1
                mpirun -np 4 python ex31p.py -m ../data/amr-hex.mesh -rs 1
'''
from mpi4py import MPI
import mfem.par as mfem
from mfem.par import intArray
import os
from os.path import expanduser, join
import numpy as np
from numpy import sin, cos, array, pi, sqrt

sqrt1_2 = 1/sqrt(2)
sqrt2 = sqrt(2)

num_procs = MPI.COMM_WORLD.size
myid = MPI.COMM_WORLD.rank
smyid = '.'+'{:0>6d}'.format(myid)


def run(order=1,
        rs=2,
        rp=1,
        freq=1,
        meshfile='',
        visualization=False,
        numba=False):

    # 3. Read the (serial) mesh from the given mesh file on all processors.  We
    #    can handle triangular, quadrilateral, tetrahedral, hexahedral, surface
    #    and volume meshes with the same code.
    mesh = mfem.Mesh(meshfile, 1, 1)
    dim = mesh.Dimension()
    sdim = mesh.SpaceDimension()

    # 4. Refine the serial mesh on all processors to increase the resolution. In
    #    this example we do 'ref_levels' of uniform refinement (2 by default, or
    #    specified on the command line with -rs).
    for x in range(rs):
        mesh.UniformRefinement()

    # 5. Define a parallel mesh by a partitioning of the serial mesh. Refine
    #    this mesh further in parallel to increase the resolution (1 time by
    #    default, or specified on the command line with -rp). Once the parallel
    #    mesh is defined, the serial mesh can be deleted.
    pmesh = mfem.ParMesh(MPI.COMM_WORLD, mesh)
    del mesh
    for lev in range(rp):
        pmesh.UniformRefinement()

    # 6. Define a parallel finite element space on the parallel mesh. Here we
    #    use the Nedelec finite elements of the specified order.
    if dim == 1:
        fec = mfem.ND_R1D_FECollection(order, dim)
    elif dim == 2:
        fec = mfem.ND_R2D_FECollection(order, dim)
    else:
        fec = mfem.ND_FECollection(order, dim)

    fespace = mfem.ParFiniteElementSpace(pmesh, fec)
    size = fespace.GlobalTrueVSize()
    if myid == 0:
        print("Number of H(curl) unknowns: " + str(size))

    # 7. Determine the list of true (i.e. parallel conforming) essential
    #    boundary dofs. In this example, the boundary conditions are defined
    #    by marking all the boundary attributes from the mesh as essential
    #    (Dirichlet) and converting them to a list of true dofs.
    ess_tdof_list = intArray()
    if pmesh.bdr_attributes.Size():
        ess_bdr = intArray([1]*pmesh.bdr_attributes.Max())
        fespace.GetEssentialTrueDofs(ess_bdr, ess_tdof_list)

    # 7.b Define coefficent to use. Here, we define outside this function.
    #     We call decoratr function (mfem.jit.xxx) manually. Also, note
    #     we are passing dim as params, so that numba uses proper dim
    #     when compiling it.
    if numba:
        params = {"dim": dim, "kappa": pi*freq}
        E = mfem.jit.vector(vdim=3, params=params)(E_exact)
        CurlE = mfem.jit.vector(vdim=3, params=params)(CurlE_exact)
        f = mfem.jit.vector(vdim=3, params=params)(f_exact)
    else:
        pass  # ToDo provide regular example.

    # 8. Set up the parallel linear form b(.) which corresponds to the
    #    right-hand side of the FEM linear system, which in this case is
    #    (f,phi_i) where f is given by the function f_exact and phi_i are the
    #    basis functions in the finite element fespace.
    b = mfem.ParLinearForm(fespace)
    b.AddDomainIntegrator(mfem.VectorFEDomainLFIntegrator(f))
    b.Assemble()

    # 9. Define the solution vector x as a parallel finite element grid function
    #    corresponding to fespace. Initialize x by projecting the exact
    #    solution. Note that only values from the boundary edges will be used
    #    when eliminating the non-homogeneous boundary condition to modify the
    #    r.h.s. vector b.
    sol = mfem.ParGridFunction(fespace)
    sol.ProjectCoefficient(E)

    # 10. Set up the parallel bilinear form corresponding to the EM diffusion
    #     operator curl muinv curl + sigma I, by adding the curl-curl and the
    #     mass domain integrators.
    mat = array([[2.0, sqrt1_2, 0, ],
                 [sqrt1_2, 2.0, sqrt1_2],
                 [0.0, sqrt1_2, 2.0, ], ])
    sigmaMat = mfem.DenseMatrix(mat)

    muinv = mfem.ConstantCoefficient(1.0)
    sigma = mfem. MatrixConstantCoefficient(sigmaMat)
    a = mfem.ParBilinearForm(fespace)
    a.AddDomainIntegrator(mfem.CurlCurlIntegrator(muinv))
    a.AddDomainIntegrator(mfem.VectorFEMassIntegrator(sigma))

    # 11. Assemble the parallel bilinear form and the corresponding linear
    #     system, applying any necessary transformations such as: parallel
    #     assembly, eliminating boundary conditions, applying conforming
    # b    constraints for non-conforming AMR, etc.
    a.Assemble()

    A = mfem.OperatorPtr()
    B = mfem.Vector()
    X = mfem.Vector()
    a.FormLinearSystem(ess_tdof_list, sol, b, A, X, B)

    # 12. Solve the system AX=B using PCG with the AMS preconditioner from hypre
    AM = A.AsHypreParMatrix()
    if myid == 0:
        print("Size of linear system: " + str(AM.GetGlobalNumRows()))

    ams = mfem.HypreAMS(AM, fespace)

    pcg = mfem.HyprePCG(AM)
    pcg.SetTol(1e-12)
    pcg.SetMaxIter(1000)
    pcg.SetPrintLevel(2)
    pcg.SetPreconditioner(ams)
    pcg.Mult(B, X)

    # 13. Recover the parallel grid function corresponding to X. This is the
    #     local finite element solution on each processor.
    a.RecoverFEMSolution(X, b, sol)

    # 14. Compute and print the CurlE norm of the error.
    error = sol.ComputeHCurlError(E, CurlE)
    if myid == 0:
        print("|| E_h - E ||_{Hcurl} = " + "{:g}".format(error)+"\n")

    # 15. Save the refined mesh and the solution in parallel. This output can
    #     be viewed later using GLVis: "glvis -np <np> -m mesh -g sol".
    smyid = '{:0>6d}'.format(myid)
    pmesh.Print('mesh.'+smyid, 8)
    sol.Save('sol.'+smyid, 8)

    # 15. Send the solution by socket to a GLVis server.
    if visualization:

        solCoef = mfem.VectorGridFunctionCoefficient(sol)
        dsolCoef = mfem.CurlGridFunctionCoefficient(sol)

        if dim == 1:
            x_sock = make_socketstrema()
            y_sock = make_socketstrema()
            z_sock = make_socketstrema()
            dy_sock = make_socketstrema()
            dz_sock = make_socketstrema()

            x_sock.precision(8)
            y_sock.precision(8)
            z_sock.precision(8)
            dy_sock.precision(8)
            dz_sock.precision(8)

            xVec = mfem.Vector([1, 0, 0.])
            yVec = mfem.Vector([0, 1, 0.])
            zVec = mfem.Vector([0, 0, 1.])

            xVecCoef = mfem.VectorConstantCoefficient(xVec)
            yVecCoef = mfem.VectorConstantCoefficient(yVec)
            zVecCoef = mfem.VectorConstantCoefficient(zVec)

            fec_h1 = mfem.H1_FECollection(order, dim)
            fec_l2 = mfem.L2_FECollection(order-1, dim)

            fes_h1 = mfem.ParFiniteElementSpace(pmesh, fec_h1)
            fes_l2 = mfem.ParFiniteElementSpace(pmesh, fec_l2)

            xComp = mfem.ParGridFunction(fes_l2)
            yComp = mfem.ParGridFunction(fes_h1)
            zComp = mfem.ParGridFunction(fes_h1)

            dyComp = mfem.ParGridFunction(fes_l2)
            dzComp = mfem.ParGridFunction(fes_l2)

            xCoef = mfem.InnerProductCoefficient(xVecCoef, solCoef)
            yCoef = mfem.InnerProductCoefficient(yVecCoef, solCoef)
            zCoef = mfem.InnerProductCoefficient(zVecCoef, solCoef)

            xComp.ProjectCoefficient(xCoef)
            yComp.ProjectCoefficient(yCoef)
            zComp.ProjectCoefficient(zCoef)

            x_sock << "solution\n" << pmesh << xComp
            x_sock.flush()
            x_sock << "window_title 'X component'"
            x_sock.flush()
            y_sock << "solution\n" << pmesh << yComp
            y_sock.flush()
            y_sock << "window_geometry 403 0 400 350 " << "window_title 'Y component'"
            y_sock.flush()
            z_sock << "solution\n" << pmesh << zComp
            z_sock.flush()
            z_sock << "window_geometry 806 0 400 350 " << "window_title 'Z component'"
            z_sock.flush()

            dyCoef = mfem.InnerProductCoefficient(yVecCoef, dsolCoef)
            dzCoef = mfem.InnerProductCoefficient(zVecCoef, dsolCoef)
            dyComp.ProjectCoefficient(dyCoef)
            dzComp.ProjectCoefficient(dzCoef)

            dy_sock << "solution\n" << pmesh << dyComp
            dy_sock.flush()
            dy_sock << "window_geometry 403 375 400 350 " << "window_title 'Y component of Curl'"
            dy_sock.flush()

            dy_sock << "solution\n" << pmesh << dzComp
            dy_sock.flush()
            dy_sock << "window_geometry 403 375 400 350 " << "window_title 'Z component of Curl'"
            dy_sock.flush()

        elif dim == 2:
            xy_sock = make_socketstrema()
            z_sock = make_socketstrema()
            dxy_sock = make_socketstrema()
            dz_sock = make_socketstrema()

            mat = array([[1.0, 0.0, 0.0], [0.0, 1.0, 0.0]])
            xyMat = mfem.DenseMatrix(mat)
            xyMatCoef = mfem.MatrixConstantCoefficient(xyMat)
            zVec = array([0, 0, 1])
            zVecCoef = mfem.VectorConstantCoefficient(zVec)

            xyCoef = mfem.MatrixVectorProductCoefficient(xyMatCoef, solCoef)
            zCoef = mfem.InnerProductCoefficient(zVecCoef, solCoef)

            fec_h1 = mfem.H1_FECollection(order, dim)
            fec_nd = mfem.ND_FECollection(order, dim)
            fec_rt = mfem.RT_FECollection(order-1, dim)
            fec_l2 = mfem.L2_FECollection(order-1, dim)

            fes_h1 = mfem.ParFiniteElementSpace(pmesh, fec_h1)
            fes_nd = mfem.ParFiniteElementSpace(pmesh, fec_nd)
            fes_rt = mfem.ParFiniteElementSpace(pmesh, fec_rt)
            fes_l2 = mfem.ParFiniteElementSpace(pmesh, fec_l2)

            xyComp = mfem.ParGridFunction(fes_nd)
            zComp = mfem.ParGridFunction(fes_h1)

            dxyComp = mfem.ParGridFunction(fes_rt)
            dzComp = mfem.ParGridFunction(fes_l2)

            xyComp.ProjectCoefficient(xyCoef)
            zComp.ProjectCoefficient(zCoef)

            xy_sock.precision(8)
            xy_sock << "solution\n" << pmesh << xyComp
            xy_sock << "window_title 'XY components'\n"
            xy_sock.flush()
            z_sock << "solution\n" << pmesh << zComp
            z_sock << "window_geometry 403 0 400 350 " << "window_title 'Z component'"
            z_sock.flush()

            dxyCoef = mfem.MatrixVectorProductCoefficient(xyMatCoef, dsolCoef)
            dzCoef = mfem.InnerProductCoefficient(zVecCoef, dsolCoef)

            dxyComp.ProjectCoefficient(dxyCoef)
            dzComp.ProjectCoefficient(dzCoef)

            dxy_sock << "solution\n" << pmesh << dxyComp
            dxy_sock << "window_geometry 0 375 400 350 " << "window_title 'XY components of Curl'"
            dxy_sock.flush()
            dz_sock << "solution\n" << pmesh << dzComp
            dz_sock << "window_geometry 403 375 400 350 " << "window_title 'Z component of Curl'"
            dz_sock.flush()

        else:
            sol_sock = make_socketstrema()
            dsol_sock = make_socketstrema()

            fec_rt = mfem.RT_FECollection(order-1, dim)
            fes_rt = mfem.ParFiniteElementSpace(pmesh, fec_rt)

            dsol = mfem.ParGridFunction(fes_rt)

            dsol.ProjectCoefficient(dsolCoef)

            sol_sock.precision(8)
            dsol_sock.precision(8)
            sol_sock << "solution\n" << pmesh << sol << "window_title 'Solution'"
            sol_sock.flush()
            dsol_sock << "solution\n" << pmesh << dsol
            dsol_sock.flush()
            dsol_sock << "window_geometry 0 375 400 350 " << "window_title 'Curl of solution'"
            dsol_sock.flush()


def make_socketstrema():
    sock = mfem.socketstream("localhost", 19916)
    sock.precision(8)
    sock << "parallel " << num_procs << " " << myid << "\n"
    return sock


def E_exact(x):
    E = np.zeros(vdim)
    if dim == 1:
        E[0] = 1.1 * sin(kappa * x[0] + 0.0 * pi)
        E[1] = 1.2 * sin(kappa * x[0] + 0.4 * pi)
        E[2] = 1.3 * sin(kappa * x[0] + 0.9 * pi)
    elif dim == 2:
        E[0] = 1.1 * sin(kappa * sqrt1_2 * (x[0] + x[1]) + 0.0 * pi)
        E[1] = 1.2 * sin(kappa * sqrt1_2 * (x[0] + x[1]) + 0.4 * pi)
        E[2] = 1.3 * sin(kappa * sqrt1_2 * (x[0] + x[1]) + 0.9 * pi)
    else:
        E[0] = 1.1 * sin(kappa * sqrt1_2 * (x[0] + x[1]) + 0.0 * pi)
        E[1] = 1.2 * sin(kappa * sqrt1_2 * (x[0] + x[1]) + 0.4 * pi)
        E[2] = 1.3 * sin(kappa * sqrt1_2 * (x[0] + x[1]) + 0.9 * pi)

        for i in range(3):
            E[i] = E[i]*cos(kappa * x[2])
    return E


def CurlE_exact(x):
    dE = np.zeros(vdim)
    if dim == 1:
        c4 = cos(kappa * x[0] + 0.4 * pi)
        c9 = cos(kappa * x[0] + 0.9 * pi)

        dE[0] = 0.0
        dE[1] = -1.3 * c9
        dE[2] = 1.2 * c4
        for i in range(3):
            dE[i] = dE[i]*kappa

    elif dim == 2:
        c0 = cos(kappa * sqrt1_2 * (x[0] + x[1]) + 0.0 * pi)
        c4 = cos(kappa * sqrt1_2 * (x[0] + x[1]) + 0.4 * pi)
        c9 = cos(kappa * sqrt1_2 * (x[0] + x[1]) + 0.9 * pi)

        dE[0] = 1.3 * c9
        dE[1] = -1.3 * c9
        dE[2] = 1.2 * c4 - 1.1 * c0
        for i in range(3):
            dE[i] = dE[i]*kappa*sqrt1_2

    else:
        s0 = sin(kappa * sqrt1_2 * (x[0] + x[1]) + 0.0 * pi)
        c0 = cos(kappa * sqrt1_2 * (x[0] + x[1]) + 0.0 * pi)
        s4 = sin(kappa * sqrt1_2 * (x[0] + x[1]) + 0.4 * pi)
        c4 = cos(kappa * sqrt1_2 * (x[0] + x[1]) + 0.4 * pi)
        c9 = cos(kappa * sqrt1_2 * (x[0] + x[1]) + 0.9 * pi)
        sk = sin(kappa * x[2])
        ck = cos(kappa * x[2])

        dE[0] = 1.2 * s4 * sk + 1.3 * sqrt1_2 * c9 * ck
        dE[1] = -1.1 * s0 * sk - 1.3 * sqrt1_2 * c9 * ck
        dE[2] = -sqrt1_2 * (1.1 * c0 - 1.2 * c4) * ck
        for i in range(3):
            dE[i] = dE[i]*kappa
    return dE


def f_exact(x):
    f = np.zeros(vdim, dtype=np.float)
    if dim == 1:
        s0 = sin(kappa * x[0] + 0.0 * pi)
        s4 = sin(kappa * x[0] + 0.4 * pi)
        s9 = sin(kappa * x[0] + 0.9 * pi)

        f[0] = 2.2 * s0 + 1.2 * sqrt1_2 * s4
        f[1] = (1.2 * (2.0 + kappa * kappa) * s4 +
                sqrt1_2 * (1.1 * s0 + 1.3 * s9))
        f[2] = 1.3 * (2.0 + kappa * kappa) * s9 + 1.2 * sqrt1_2 * s4

    elif dim == 2:
        s0 = sin(kappa * sqrt1_2 * (x[0] + x[1]) + 0.0 * pi)
        s4 = sin(kappa * sqrt1_2 * (x[0] + x[1]) + 0.4 * pi)
        s9 = sin(kappa * sqrt1_2 * (x[0] + x[1]) + 0.9 * pi)

        f[0] = (0.55 * (4.0 + kappa * kappa) * s0 +
                0.6 * (sqrt2 - kappa * kappa) * s4)
        f[1] = (0.55 * (sqrt2 - kappa * kappa) * s0 +
                0.6 * (4.0 + kappa * kappa) * s4 +
                0.65 * sqrt2 * s9)
        f[2] = 0.6 * sqrt2 * s4 + 1.3 * (2.0 + kappa * kappa) * s9

    else:
        s0 = sin(kappa * sqrt1_2 * (x[0] + x[1]) + 0.0 * pi)
        c0 = cos(kappa * sqrt1_2 * (x[0] + x[1]) + 0.0 * pi)
        s4 = sin(kappa * sqrt1_2 * (x[0] + x[1]) + 0.4 * pi)
        c4 = cos(kappa * sqrt1_2 * (x[0] + x[1]) + 0.4 * pi)
        s9 = sin(kappa * sqrt1_2 * (x[0] + x[1]) + 0.9 * pi)
        c9 = cos(kappa * sqrt1_2 * (x[0] + x[1]) + 0.9 * pi)
        sk = sin(kappa * x[2])
        ck = cos(kappa * x[2])

        f[0] = (0.55 * (4.0 + 3.0 * kappa * kappa) * s0 * ck +
                0.6 * (sqrt2 - kappa * kappa) * s4 * ck -
                0.65 * sqrt2 * kappa * kappa * c9 * sk)

        f[1] = (0.55 * (sqrt2 - kappa * kappa) * s0 * ck +
                0.6 * (4.0 + 3.0 * kappa * kappa) * s4 * ck +
                0.65 * sqrt2 * s9 * ck -
                0.65 * sqrt2 * kappa * kappa * c9 * sk)

        f[2] = (0.6 * sqrt2 * s4 * ck -
                sqrt2 * kappa * kappa * (0.55 * c0 + 0.6 * c4) * sk
                + 1.3 * (2.0 + kappa * kappa) * s9 * ck)
    return f


if __name__ == "__main__":
    from mfem.common.arg_parser import ArgParser

    parser = ArgParser(description='Ex31 (Definite Maxwell Problem)')
    parser.add_argument('-m', '--mesh',
                        default="inline-quad.mesh",
                        action='store', type=str,
                        help='Mesh file to use.')
    parser.add_argument('-rs', '--refine-serial',
                        action='store', default=2, type=int,
                        help="Number of times to refine the mesh uniformly in serial")
    parser.add_argument('-rp', '--refine-parallel',
                        action='store', default=1, type=int,
                        help="Number of times to refine the mesh uniformly in paralle.")
    parser.add_argument('-o', '--order',
                        action='store', default=1, type=int,
                        help="Finite element order (polynomial degree)")
    parser.add_argument("-f", "--frequency",
                        action='store',
                        type=float,
                        default=1.0,
                        help="Set the frequency for the exact")
    parser.add_argument('-vis', '--visualization',
                        action='store_true',
                        help='Enable GLVis visualization')

    try:
        from numba import jit
        HAS_NUMBA = True
    except ImportError:
        assert False, "This example requires numba to run"
    parser.add_argument("-n", "--numba",
                        default=int(HAS_NUMBA),
                        type=int,
                        help="Use Number compiled coefficient")

    args = parser.parse_args()
    args.numba = bool(args.numba)
    if myid == 0:
        parser.print_options(args)

    order = args.order

    meshfile = expanduser(
        join(os.path.dirname(__file__), '..', 'data', args.mesh))
    visualization = args.visualization
    freq = args.frequency
    numba = args.numba

    run(freq=freq,
        order=order,
        rs=args.refine_serial,
        rp=args.refine_parallel,
        meshfile=meshfile,
        visualization=visualization,
        numba=numba)



'''
   MFEM example 18
      This is a version of Example 18 with a simple adaptive mesh
      refinement loop. 
      See c++ version in the MFEM library for more detail 
'''
from ex18_common import FE_Evolution, InitialCondition, RiemannSolver, DomainIntegrator, FaceIntegrator
from mfem.common.arg_parser import ArgParser
import mfem.ser as mfem
from mfem.ser import intArray
from os.path import expanduser, join, dirname
import numpy as np
from numpy import sqrt, pi, cos, sin, hypot, arctan2
from scipy.special import erfc

# Equation constant parameters.(using globals to share them with ex18_common)
import ex18_common


def run(problem=1,
        ref_levels=1,
        order=3,
        ode_solver_type=4,
        t_final=0.5,
        dt=-0.01,
        cfl=0.3,
        visualization=True,
        vis_steps=50,
        meshfile=''):

    ex18_common.num_equation = 4
    ex18_common.specific_heat_ratio = 1.4
    ex18_common.gas_constant = 1.0
    ex18_common.problem = problem
    num_equation = ex18_common.num_equation

    # 2. Read the mesh from the given mesh file. This example requires a 2D
    #    periodic mesh, such as ../data/periodic-square.mesh.
    meshfile = expanduser(join(dirname(__file__), '..', 'data', meshfile))
    mesh = mfem.Mesh(meshfile, 1, 1)
    dim = mesh.Dimension()

    # 3. Define the ODE solver used for time integration. Several explicit
    #    Runge-Kutta methods are available.
    ode_solver = None
    if ode_solver_type == 1:
        ode_solver = mfem.ForwardEulerSolver()
    elif ode_solver_type == 2:
        ode_solver = mfem.RK2Solver(1.0)
    elif ode_solver_type == 3:
        ode_solver = mfem.RK3SSolver()
    elif ode_solver_type == 4:
        ode_solver = mfem.RK4Solver()
    elif ode_solver_type == 6:
        ode_solver = mfem.RK6Solver()
    else:
        print("Unknown ODE solver type: " + str(ode_solver_type))
        exit

    # 4. Refine the mesh to increase the resolution. In this example we do
    #    'ref_levels' of uniform refinement, where 'ref_levels' is a
    #    command-line parameter.
    for lev in range(ref_levels):
        mesh.UniformRefinement()

    # 5. Define the discontinuous DG finite element space of the given
    #    polynomial order on the refined mesh.

    fec = mfem.DG_FECollection(order, dim)
    # Finite element space for a scalar (thermodynamic quantity)
    fes = mfem.FiniteElementSpace(mesh, fec)
    # Finite element space for a mesh-dim vector quantity (momentum)
    dfes = mfem.FiniteElementSpace(mesh, fec, dim, mfem.Ordering.byNODES)
    # Finite element space for all variables together (total thermodynamic state)
    vfes = mfem.FiniteElementSpace(
        mesh, fec, num_equation, mfem.Ordering.byNODES)

    assert fes.GetOrdering() == mfem.Ordering.byNODES, "Ordering must be byNODES"
    print("Number of unknowns: " + str(vfes.GetVSize()))

    # 6. Define the initial conditions, save the corresponding mesh and grid
    #    functions to a file. This can be opened with GLVis with the -gc option.
    #    The solution u has components {density, x-momentum, y-momentum, energy}.
    #    These are stored contiguously in the BlockVector u_block.

    offsets = [k*vfes.GetNDofs() for k in range(num_equation+1)]
    offsets = mfem.intArray(offsets)
    u_block = mfem.BlockVector(offsets)
    mom = mfem.GridFunction(dfes, u_block,  offsets[1])

    #
    #  Define coefficient using VecotrPyCoefficient and PyCoefficient
    #  A user needs to define EvalValue method
    #
    u0 = InitialCondition(num_equation)
    sol = mfem.GridFunction(vfes, u_block.GetData())
    sol.ProjectCoefficient(u0)

    mesh.Print("vortex.mesh", 8)
    for k in range(num_equation):
        uk = mfem.GridFunction(fes, u_block.GetBlock(k).GetData())
        sol_name = "vortex-" + str(k) + "-init.gf"
        uk.Save(sol_name, 8)

    #  7. Set up the nonlinear form corresponding to the DG discretization of the
    #     flux divergence, and assemble the corresponding mass matrix.
    Aflux = mfem.MixedBilinearForm(dfes, fes)
    Aflux.AddDomainIntegrator(DomainIntegrator(dim))
    Aflux.Assemble()

    A = mfem.NonlinearForm(vfes)
    rsolver = RiemannSolver()
    ii = FaceIntegrator(rsolver, dim)
    A.AddInteriorFaceIntegrator(ii)

    #  8. Define the time-dependent evolution operator describing the ODE
    #     right-hand side, and perform time-integration (looping over the time
    #     iterations, ti, with a time-step dt).
    euler = FE_Evolution(vfes, A, Aflux.SpMat())

    if (visualization):
        sout = mfem.socketstream("localhost", 19916)
        sout.precision(8)
        sout << "solution\n" << mesh << mom
        sout << "pause\n"
        sout.flush()
        print("GLVis visualization paused.")
        print(" Press space (in the GLVis window) to resume it.")

    # Determine the minimum element size.
    hmin = 0
    if (cfl > 0):
        hmin = min([mesh.GetElementSize(i, 1) for i in range(mesh.GetNE())])

    t = 0.0
    euler.SetTime(t)
    ode_solver.Init(euler)
    if (cfl > 0):
        #  Find a safe dt, using a temporary vector. Calling Mult() computes the
        #  maximum char speed at all quadrature points on all faces.
        z = mfem.Vector(A.Width())
        A.Mult(sol, z)

        dt = cfl * hmin / ex18_common.max_char_speed / (2*order+1)

    # Integrate in time.
    done = False
    ti = 0
    while not done:
        dt_real = min(dt, t_final - t)

        t, dt_real = ode_solver.Step(sol, t, dt_real)

        if (cfl > 0):
            dt = cfl * hmin / ex18_common.max_char_speed / (2*order+1)
        ti = ti+1
        done = (t >= t_final - 1e-8*dt)
        if (done or ti % vis_steps == 0):
            print("time step: " + str(ti) + ", time: " + "{:g}".format(t))
            if (visualization):
                sout << "solution\n" << mesh << mom << flush

    #  9. Save the final solution. This output can be viewed later using GLVis:
    #    "glvis -m vortex.mesh -g vortex-1-final.gf".
    for k in range(num_equation):
        uk = mfem.GridFunction(fes, u_block.GetBlock(k).GetData())
        sol_name = "vortex-" + str(k) + "-final.gf"
        uk.Save(sol_name, 8)

    print(" done")
    # 10. Compute the L2 solution error summed for all components.
    if (t_final == 2.0):
        error = sol.ComputeLpError(2., u0)
        print("Solution error: " + str(error))


if __name__ == "__main__":

    parser = ArgParser(description='Ex18')
    parser.add_argument('-m', '--mesh',
                        default='periodic-square.mesh',
                        action='store', type=str,
                        help='Mesh file to use.')
    parser.add_argument('-p', '--problem',
                        action='store', default=1, type=int,
                        help='Problem setup to use. See options in velocity_function().')
    parser.add_argument('-r', '--refine',
                        action='store', default=1, type=int,
                        help="Number of times to refine the mesh uniformly.")
    parser.add_argument('-o', '--order',
                        action='store', default=3, type=int,
                        help="Finite element order (polynomial degree)")
    parser.add_argument('-s', '--ode_solver',
                        action='store', default=4, type=int,
                        help="ODE solver: 1 - Forward Euler,\n\t" +
                        "            2 - RK2 SSP, 3 - RK3 SSP, 4 - RK4, 6 - RK6.")
    parser.add_argument('-tf', '--t_final',
                        action='store', default=2.0, type=float,
                        help="Final time; start time is 0.")
    parser.add_argument("-dt", "--time_step",
                        action='store', default=-0.01, type=float,
                        help="Time step.")
    parser.add_argument('-c', '--cfl_number',
                        action='store', default=0.3, type=float,
                        help="CFL number for timestep calculation.")
    parser.add_argument('-vis', '--visualization',
                        action='store_true',
                        help='Enable GLVis visualization')
    parser.add_argument('-vs', '--visualization-steps',
                        action='store', default=50, type=float,
                        help="Visualize every n-th timestep.")

    args = parser.parse_args()

    parser.print_options(args)

    run(problem=args.problem,
        ref_levels=args.refine,
        order=args.order,
        ode_solver_type=args.ode_solver,
        t_final=args.t_final,
        dt=args.time_step,
        cfl=args.cfl_number,
        visualization=args.visualization,
        vis_steps=args.visualization_steps,
        meshfile=args.mesh)



'''
   MFEM example 6
      This is a version of Example 1 with a simple adaptive mesh
      refinement loop. 
      See c++ version in the MFEM library for more detail 
'''
import mfem.par as mfem
from mfem.par import intArray
from os.path import expanduser, join, dirname
from mpi4py import MPI
import numpy as np
import os

num_proc = MPI.COMM_WORLD.size
myid = MPI.COMM_WORLD.rank
smyid = '{:0>6d}'.format(myid)
verbose = (myid == 0)


def run(meshfile="",
        order=1,
        smooth_el=True,
        max_dofs=100000,
        reorder_mesh=0,
        restart=False,
        nc_simplices=True,
        visualization=False,
        device_config="cpu",
        pa=False):

    device = mfem.Device(device_config)
    if myid == 0:
        device.Print()

    # 4. Read the (serial) mesh from the given mesh file on all processors.
    #    We can handle triangular, quadrilateral, tetrahedral, hexahedral,
    #    surface and volume meshes with the same code.

    mesh = mfem.Mesh(meshfile, 1, 1)

    # 5. A NURBS mesh cannot be refined locally so we refine it uniformly
    #    and project it to a standard curvilinear mesh of order 2.
    if (mesh.NURBSext):
        mesh.UniformRefinement()
        mesh.SetCurvature(2)

    # 6. MFEM supports dynamic partitioning (load balancing) of parallel non-
    #    conforming meshes based on space-filling curve (SFC) partitioning.
    #    SFC partitioning is extremely fast and scales to hundreds of
    #    thousands of processors, but requires the coarse mesh to be ordered,
    #    ideally as a sequence of face-neighbors. The mesh may already be
    #    ordered (like star-hilbert.mesh) or we can order it here. Ordering
    #    type 1 is a fast spatial sort of the mesh, type 2 is a high quality
    #    optimization algorithm suitable for ordering general unstructured
    #    meshes.
    print("reorder_mesh", reorder_mesh)
    if reorder_mesh > 0:
        ordering = mfem.intArray()
        if reorder_mesh == 1:
            mesh.GetHilbertElementOrdering(ordering)
        elif reorder_mesh == 2:
            mesh.GetGeckoElementOrdering(ordering)
        mesh.ReorderElements(ordering)

    # 7. Make sure the mesh is in the non-conforming mode to enable local
    #    refinement of quadrilaterals/hexahedra, and the above partitioning
    #    algorithm. Simplices can be refined either in conforming or in non-
    #    conforming mode. The conforming mode however does not support
    #    dynamic partitioning.
    mesh.EnsureNCMesh(nc_simplices)

    # 8. Define a parallel mesh by partitioning the serial mesh.
    #    Once the parallel mesh is defined, the serial mesh can be deleted.
    pmesh = mfem.ParMesh(MPI.COMM_WORLD, mesh)

    dim = pmesh.Dimension()
    sdim = pmesh.SpaceDimension()

    ess_bdr = intArray(pmesh.bdr_attributes.Max())
    ess_bdr.Assign(1)

    # 10. Define a finite element space on the mesh. The polynomial order is
    #     one (linear) by default, but this can be changed on the command line.
    fec = mfem.H1_FECollection(order, dim)
    fespace = mfem.ParFiniteElementSpace(pmesh, fec)

    # 11. As in Example 1p, we set up bilinear and linear forms corresponding to
    #     the Laplace problem -\Delta u = 1. We don't assemble the discrete
    #     problem yet, this will be done in the main loop.
    a = mfem.ParBilinearForm(fespace)
    if pa:
        a.SetAssemblyLevel(mfem.AssemblyLevel_PARTIAL)
        a.SetDiagonalPolicy(mfem.Operator.DIAG_ONE)
    b = mfem.ParLinearForm(fespace)

    one = mfem.ConstantCoefficient(1.0)

    integ = mfem.DiffusionIntegrator(one)
    a.AddDomainIntegrator(integ)
    b.AddDomainIntegrator(mfem.DomainLFIntegrator(one))

    # 12. The solution vector x and the associated finite element grid function
    #     will be maintained over the AMR iterations. We initialize it to zero.
    x = mfem.ParGridFunction(fespace)
    x.Assign(0.0)

    # 13. Connect to GLVis.
    if visualization:
        sout = mfem.socketstream('localhost', 19916)
        if sout.good():
            sout.precision(8)
        else:
            visualization = False

    # 14. Set up an error estimator. Here we use the Zienkiewicz-Zhu estimator
    #     with L2 projection in the smoothing step to better handle hanging
    #     nodes and parallel partitioning. We need to supply a space for the
    #     discontinuous flux (L2) and a space for the smoothed flux

    flux_fec = mfem.L2_FECollection(order, dim)
    flux_fes = mfem.ParFiniteElementSpace(pmesh, flux_fec, sdim)

    if smooth_el and dim > 1:
        smooth_flux_fec = mfem.RT_FECollection(order-1, dim)
        smooth_flux_fes = mfem.ParFiniteElementSpace(pmesh, smooth_flux_fec, 1)
    else:
        smooth_flux_fec = mfem.H1_FECollection(order, dim)
        smooth_flux_fes = mfem.ParFiniteElementSpace(
            pmesh, smooth_flux_fec, dim)

    estimator = mfem.L2ZienkiewiczZhuEstimator(integ, x, flux_fes,
                                               smooth_flux_fes)
    # 15. A refiner selects and refines elements based on a refinement strategy.
    #     The strategy here is to refine elements with errors larger than a
    #     fraction of the maximum element error. Other strategies are possible.
    #     The refiner will call the given error estimator.
    refiner = mfem.ThresholdRefiner(estimator)
    refiner.SetTotalErrorFraction(0.7)

    # 16. The main AMR loop. In each iteration we solve the problem on the
    #     current mesh, visualize the solution, and refine the mesh.
    it = 0
    while True:
        global_dofs = fespace.GlobalTrueVSize()
        if (myid == 0):
            print("AMR iteration " + str(it))
            print("Number of unknowns: " + str(global_dofs))

        # 17. Assemble the right-hand side and determine the list of true
        #     (i.e. parallel conforming) essential boundary dofs.
        ess_tdof_list = intArray()
        fespace.GetEssentialTrueDofs(ess_bdr, ess_tdof_list)
        b.Assemble()

        # 18. Assemble the stiffness matrix. Note that MFEM doesn't care at this
        #     point that the mesh is nonconforming and parallel.  The FE space is
        #     considered 'cut' along hanging edges/faces, and also across
        #     processor boundaries.
        a.Assemble()

        # 19. Create the linear system: eliminate boundary conditions.
        #     The system will be solved for true (unconstrained) DOFs only.
        A = mfem.OperatorPtr()
        B = mfem.Vector()
        X = mfem.Vector()
        copy_interior = 1
        a.FormLinearSystem(ess_tdof_list, x, b, A, X, B, copy_interior)

        # 20. Solve the linear system A X = B.
        #     * With full assembly, use the BoomerAMG preconditioner from hypre.
        #     * With partial assembly, use a diagonal preconditioner.
        if pa:
            M = mfem.OperatorJacobiSmoother(a, ess_tdof_list)
        else:
            amg = mfem.HypreBoomerAMG()
            amg.SetPrintLevel(0)
            M = amg

        pcg = mfem.CGSolver(MPI.COMM_WORLD)
        pcg.SetPreconditioner(M)
        pcg.SetOperator(A.Ptr())
        pcg.SetRelTol(1e-6)
        pcg.SetMaxIter(2000)
        pcg.SetPrintLevel(3)
        pcg.Mult(B, X)

        # 21. Switch back to the host and extract the parallel grid function
        #     corresponding to the finite element approximation X. This is the
        #     local solution on each processor.
        a.RecoverFEMSolution(X, b, x)
        if (global_dofs > max_dofs):
            if (myid == 0):
                print("Reached the maximum number of dofs. Stop.")
            break

        # 22. Send the solution by socket to a GLVis server.
        if visualization:
            sout << "parallel " << num_proc << " " << myid << "\n"
            sout << "solution\n" << pmesh << x
            sout.flush()

        # 23. Call the refiner to modify the mesh. The refiner calls the error
        #     estimator to obtain element errors, then it selects elements to be
        #     refined and finally it modifies the mesh. The Stop() method can be
        #     used to determine if a stopping criterion was met.

        refiner.Apply(pmesh)
        if (refiner.Stop()):
            if myid == 0:
                print("Stopping criterion satisfied. Stop.")
            break
        # 24. Update the finite element space (recalculate the number of DOFs,
        #     etc.) and create a grid function update matrix. Apply the matrix
        #     to any GridFunctions over the space. In this case, the update
        #     matrix is an interpolation matrix so the updated GridFunction will
        #     still represent the same function as before refinement.
        fespace.Update()
        x.Update()

        # 25. Load balance the mesh, and update the space and solution. Currently
        #     available only for nonconforming meshes.
        if pmesh.Nonconforming():
            pmesh.Rebalance()
            # Update the space and the GridFunction. This time the update matrix
            # redistributes the GridFunction among the processors.
            fespace.Update()
            x.Update()

        # 26. Inform also the bilinear and linear forms that the space has
        #     changed.
        a.Update()
        b.Update()

        # 27. Save the current state of the mesh every 5 iterations. The
        #     computation can be restarted from this point. Note that unlike in
        #     visualization, we need to use the 'ParPrint' method to save all
        #     internal parallel data structures.

        if ((it + 1) % 5) == 0:
            pmesh.ParPrint("ex6p-checkpoint."+smyid, 8)
            if myid == 0:
                print("Checkpoint saved")
        it = it + 1


if __name__ == "__main__":
    from mfem.common.arg_parser import ArgParser

    parser = ArgParser(
        description='Ex6p (Laplace problem with AMR)')
    parser.add_argument('-m', '--mesh',
                        default="star.mesh",
                        action='store', type=str,
                        help='Mesh file to use.')
    parser.add_argument('-o', '--order',
                        action='store',
                        default=1,
                        type=int,
                        help="Finite element order (polynomial degree)")
    parser.add_argument("-pa", "--partial-assembly",
                        action='store_true',
                        help="Enable Partial Assembly.")
    parser.add_argument("-d", "--device",
                        default="cpu", type=str,
                        help="Device configuration string, see Device::Configure().")
    parser.add_argument("-rm", "--reorder-mesh",
                        action='store',
                        default=0,
                        type=int,
                        help="".join(["Reorder elements of the coarse mesh to improve ",
                                      "dynamic partitioning: 0=none, 1=hilbert, 2=gecko."]))
    parser.add_argument("-cs", "--conforming-simplices",
                        action='store_true',
                        default=False,
                        help="For simplicial meshes, disable nonconforming refinement")
    parser.add_argument("-md", "--max-dofs",
                        action="store",
                        type=int,
                        default=100000,
                        help="Stop after reaching this many degrees of freedom.")
    parser.add_argument("-h1", "--smooth-h1",
                        action='store_true',
                        default=False,
                        help="Represent the smooth flux in vector H1 space.")
    # parser.add_argument("-res", "--restart",
    #                    action='store_true',
    #                    help="Restart computation from the last checkpoint.");

    parser.add_argument('-no-vis', '--no-visualization',
                        action='store_true',
                        default=False,
                        help='Enable GLVis visualization')

    args = parser.parse_args()
    if myid == 0:
        parser.print_options(args)

    meshfile = expanduser(
        join(os.path.dirname(__file__), '..', 'data', args.mesh))

    smooth_el = not args.smooth_h1

    run(meshfile=meshfile,
        order=args.order,
        smooth_el=smooth_el,
        max_dofs=args.max_dofs,
        reorder_mesh=args.reorder_mesh,
        restart=False,
        nc_simplices=not args.conforming_simplices,
        visualization=not args.no_visualization,
        device_config=args.device,
        pa=args.partial_assembly)



'''
   MFEM example 33

   See c++ version in the MFEM library for more detail

   Sample runs:  mpirun -np 2 ex33p.py -m ../data/square-disc.mesh -alpha 0.33 -o 2
                 mpirun -np 2 ex33p.py -m ../data/square-disc.mesh -alpha 4.5 -o 3
                 mpirun -np 2 ex33p.py -m ../data/star.mesh -alpha 1.4 -o 3
                 mpirun -np 2 ex33p.py -m ../data/star.mesh -alpha 0.99 -o 3
                 mpirun -np 2 ex33p.py -m ../data/inline-quad.mesh -alpha 0.5 -o 3
                 mpirun -np 2 ex33p.py -m ../data/amr-quad.mesh -alpha 1.5 -o 3
                 mpirun -np 2 ex33p.py -m ../data/disc-nurbs.mesh -alpha 0.33 -o 3
                 mpirun -np 2 ex33p.py -m ../data/disc-nurbs.mesh -alpha 2.4 -o 3 -r 4
                 mpirun -np 2 ex33p.py -m ../data/l-shape.mesh -alpha 0.33 -o 3 -r 4
                 mpirun -np 2 ex33p.py -m ../data/l-shape.mesh -alpha 1.7 -o 3 -r 5
    Verification runs:
                 mpirun -np 2 ex33p.py -m ../data/inline-segment.mesh -ver -alpha 1.7 -o 2 -r 2
                 mpirun -np 2 ex33p.py -m ../data/inline-quad.mesh -ver -alpha 1.2 -o 2 -r 2
                 mpirun -np 2 ex33p.py -m ../data/amr-quad.mesh -ver -alpha 2.6 -o 2 -r 2
                 mpirun -np 2 ex33p.py -m ../data/inline-hex.mesh -ver -alpha 0.3 -o 2 -r 1
'''
import mfem.par as mfem
from mfem.par import intArray, doubleArray
import os
from os.path import expanduser, join
import numpy as np
from numpy import sin, cos, array, pi, sqrt, floor

from mpi4py import MPI
num_procs = MPI.COMM_WORLD.size
myid = MPI.COMM_WORLD.rank


def run(alpha=0.5,
        order=1,
        num_refs=3,
        meshfile=None,
        visualization=False,
        verification=False,
        vef=False):

    coeffs = doubleArray()
    poles = doubleArray()
    progress_steps = 1

    power_of_laplace = floor(alpha)
    exponent_to_approximate = alpha - power_of_laplace

    if exponent_to_approximate > 1e-12:
        print("Approximating the fractional exponent " +
              str(exponent_to_approximate))

        from ex33_common import ComputePartialFractionApproximation
        poles, coeffs = ComputePartialFractionApproximation(
            exponent_to_approximate)

        # If the example is build without LAPACK, the exponent_to_approximate
        # might be modified by the function call above.
        alpha = exponent_to_approximate + power_of_laplace
        integer_order = False
    else:
        print("Treating integer order PDE.")
        integer_order = True

    mesh = mfem.Mesh(meshfile, 1, 1)
    dim = mesh.Dimension()

    # 4. Refine the mesh to increase the resolution.
    for i in range(num_refs):
        mesh.UniformRefinement()

    pmesh = mfem.ParMesh(MPI.COMM_WORLD, mesh)
    del mesh

    # 5. Define a finite element space on the mesh.
    fec = mfem.H1_FECollection(order, dim)
    fespace = mfem.ParFiniteElementSpace(pmesh, fec)

    fe_size = fespace.GlobalTrueVSize()
    if myid == 0:
        print("Number of finite element unknowns: " + str(fe_size))

    # 6. Determine the list of true (i.e. conforming) essential boundary dofs.
    ess_tdof_list = intArray()
    if pmesh.bdr_attributes.Size() > 0:
        ess_bdr = intArray([1]*pmesh.bdr_attributes.Max())
        fespace.GetEssentialTrueDofs(ess_bdr, ess_tdof_list)

    # 7. Define diffusion coefficient, load, and solution GridFunction.
    @mfem.jit.scalar
    def f(x):
        val = 1.0
        for i in range(len(x)):
            val *= sin(pi*x[i])
        return (len(x)*pi**2)**alpha * val

    one = mfem.ConstantCoefficient(1.0)
    u = mfem.ParGridFunction(fespace)
    x = mfem.ParGridFunction(fespace)
    g = mfem.ParGridFunction(fespace)

    u.Assign(0.0)
    x.Assign(0.0)
    g.Assign(0.0)

    # 9. Set up the linear form b(.) for integer-order PDE solves.
    b = mfem.ParLinearForm(fespace)
    if verification:
        # This statement is only relevant for the verification of the code. It
        # uses a different f such that an analytic solution is known and easy
        # to compare with the numerical one. The FPDE becomes:
        # (-Δ)^α u = (2\pi ^2)^α sin(\pi x) sin(\pi y) on [0,1]^2
        # -> u(x,y) = sin(\pi x) sin(\pi y)
        b.AddDomainIntegrator(mfem.DomainLFIntegrator(f))
    else:
        b.AddDomainIntegrator(mfem.DomainLFIntegrator(one))

    b.Assemble()

    # ------------------------------------------------------------------------
    # 10. Solve the PDE (-Δ)^N g = f, i.e. compute g = (-Δ)^{-1}^N f.
    # ------------------------------------------------------------------------

    if power_of_laplace > 0:
        # 10.1 Compute Stiffnes Matrix
        k = mfem.ParBilinearForm(fespace)
        k.AddDomainIntegrator(mfem.DiffusionIntegrator(one))
        k.Assemble()

        # 10.2 Compute Mass Matrix
        m = mfem.ParBilinearForm(fespace)
        m.AddDomainIntegrator(mfem.MassIntegrator(one))
        m.Assemble()

        mass = mfem.HypreParMatrix()
        empty = intArray()
        m.FormSystemMatrix(empty, mass)

        #  10.3 Form the system of equations
        B = mfem.Vector()
        X = mfem.Vector()
        Op = mfem.OperatorPtr()
        k.FormLinearSystem(ess_tdof_list, g, b, Op, X, B)

        prec = mfem.HypreBoomerAMG()
        prec.SetPrintLevel(-1)
        cg = mfem.CGSolver(MPI.COMM_WORLD)
        cg.SetRelTol(1e-12)
        cg.SetMaxIter(2000)
        cg.SetPrintLevel(3)
        cg.SetPreconditioner(prec)
        cg.SetOperator(Op.Ptr())

        if myid == 0:
            print("\n".join(["",
                             "Computing (-Δ) ^ -" + str(power_of_laplace) + " ( f ) "]))

        for i in range(power_of_laplace):
            # 10.4 Solve the linear system Op X = B (N times).
            cg.Mult(B, X)

            # 10.5 Visualize the solution g of -Δ ^ N g = f in the last step
            if i == power_of_laplace - 1:
                # Needed for visualization and solution verification.
                k.RecoverFEMSolution(X, b, g)
                if integer_order and verification:
                    # For an integer order PDE, g is also our solution u.
                    u += g
                if visualization:
                    fout = mfem.socketstream("localhost", 19916)
                    fout.precision(8)
                    fout << "parallel " << num_procs << " " << myid << "\n"
                    fout << "solution\n" << pmesh << g

                    progress_steps = progress_steps + 1
                    title = ("Step " + str(progress_steps) + ": Solution of PDE -Δ ^ " +
                             str(power_of_laplace) + " g = f")
                    fout << "window_title '" + title + "'"
                    fout.flush()

            # 10.6 Prepare for next iteration (primal / dual space)
            mass.Mult(X, B)
            X.SetSubVectorComplement(ess_tdof_list, 0.0)

        # 10.7 Extract solution for the next step. The b now corresponds to the
        #      function g in the PDE.
        rm = mfem.fespace.GetRestrictionMatrix()
        rm.MultTranspose(B, b)

    # ------------------------------------------------------------------------
    # 11. Solve the fractional PDE by solving M integer order PDEs and adding
    #     up the solutions.
    # ------------------------------------------------------------------------
    if not integer_order:
        # Setup visualization.
        if visualization:
            xout = mfem.socketstream("localhost", 19916)
            uout = mfem.socketstream("localhost", 19916)

        # Iterate over all expansion coefficient that contribute to the
        # solution.
        for i in range(len(coeffs)):
            if myid == 0:
                print("\nSolving PDE -Δ u + " + "{:g}".format(-poles[i]) +
                      " u = " + "{:g}".format(coeffs[i]) + " g ")

            # 11.1 Reset GridFunction for integer-order PDE solve.
            x.Assign(0.0)

            # 11.2 Set up the bilinear form a(.,.) for integer-order PDE solve.
            a = mfem.ParBilinearForm(fespace)
            a.AddDomainIntegrator(mfem.DiffusionIntegrator(one))
            d_i = mfem.ConstantCoefficient(-poles[i])
            a.AddDomainIntegrator(mfem.MassIntegrator(d_i))
            a.Assemble()

            # 11.3 Assemble the bilinear form and the corresponding linear system.
            Op = mfem.OperatorPtr()
            B = mfem.Vector()
            X = mfem.Vector()
            a.FormLinearSystem(ess_tdof_list, x, b, Op, X, B)

            # 11.4 Solve the linear system A X = B.
            prec = mfem.HypreBoomerAMG()
            prec.SetPrintLevel(-1)
            cg = mfem.CGSolver(MPI.COMM_WORLD)
            cg.SetRelTol(1e-12)
            cg.SetMaxIter(2000)
            cg.SetPrintLevel(3)
            cg.SetPreconditioner(prec)
            cg.SetOperator(Op.Ptr())
            cg.Mult(B, X)

            # 11.5 Recover the solution as a finite element grid function.
            a.RecoverFEMSolution(X, b, x)

            # 11.6 Accumulate integer-order PDE solutions.
            x *= coeffs[i]
            u += x

            # 11.7 Send fractional PDE solution to a GLVis server.
            if visualization:
                oss_x = ("Step " + str(progress_steps) +
                         ": Solution of PDE -Δ u + " + "{:g}".format(-poles[i]) +
                         " u = " + "{:g}".format(coeffs[i]) + " g")
                xout << "parallel " << num_procs << " " << myid << "\n"
                xout << "solution\n" << pmesh << x
                xout << "window_title '" << oss_x << "'"
                xout.flush()

                oss_x = ("Step " + str(progress_steps+1) +
                         ": Solution of fractional PDE (-Δ)^ " + "{:g}".format(alpha) +
                         " u = f")
                uout << "parallel " << num_procs << " " << myid << "\n"
                uout << "solution\n" << pmesh << u
                uout << "window_title '" << oss_x << "'"
                uout.flush()

    # ------------------------------------------------------------------------
    # 12. (optional) Verify the solution.
    # ------------------------------------------------------------------------
    if verification:
        @mfem.jit.scalar
        def sol(x):
            val = 1.0
            for i in range(len(x)):
                val *= sin(pi*x[i])
            return val

        l2_error = u.ComputeL2Error(sol)

        if myid == 0:
            if dim == 1:
                analytic_solution = "sin(π x)"
                expected_mesh = "inline_segment.mesh"
            elif dim == 2:
                analytic_solution = "sin(π x) sin(π y)"
                expected_mesh = "inline_quad.mesh"
            else:
                analytic_solution = "sin(π x) sin(π y) sin(π z)"
                expected_mesh = "inline_hex.mesh"

            print("\n" + "="*80 +
                  "\n\nSolution Verification in " + str(dim) + "D \n\n" +
                  "Analytic solution : " + analytic_solution + "\n" +
                  "Expected mesh     : " + expected_mesh + "\n" +
                  "Your mesh         : " + meshfile + "\n" +
                  "L2 error          : " + "{:g}".format(l2_error) + "\n\n" +
                  "="*80)


if __name__ == "__main__":
    from mfem.common.arg_parser import ArgParser

    parser = ArgParser(description='Ex33 (Fractional PDE)')
    parser.add_argument('-m', '--mesh',
                        default="star.mesh",
                        action='store', type=str,
                        help='Mesh file to use.')
    parser.add_argument('-o', '--order',
                        action='store', default=1, type=int,
                        help="Finite element order (polynomial degree) or -1 for isoparametric space")
    parser.add_argument('-r', '--refs',
                        action='store', default=3, type=int,
                        help="Number of times to refine the mesh uniformly in serial")
    parser.add_argument("-alpha", "--alpha",
                        action='store',
                        type=float,
                        default=0.5,
                        help="Fractional exponent")
    parser.add_argument('-no-vis', '--no-visualization',
                        action='store_true',
                        help='Disable GLVis visualization')
    parser.add_argument('-ver', '--verification',
                        action='store_true',
                        help="Use sinusoidal function (f) for analytic comparison.")

    try:
        from numba import jit
        HAS_NUMBA = True
    except ImportError:
        assert False, "This example requires numba to run"

    args = parser.parse_args()
    parser.print_options(args)

    meshfile = expanduser(
        join(os.path.dirname(__file__), '..', 'data', args.mesh))

    visualization = False if args.no_visualization else True

    run(alpha=args.alpha,
        order=args.order,
        num_refs=args.refs,
        meshfile=meshfile,
        verification=args.verification,
        visualization=visualization,
        vef=args.verification)



'''
   MFEM example 20p
      See c++ version in the MFEM library for more detail 
'''
import os
import mfem.par as mfem
from mfem.par import intArray
from os.path import expanduser, join, dirname
import numpy as np
from numpy import sin, cos, exp, sqrt, pi

from mpi4py import MPI
num_procs = MPI.COMM_WORLD.size
myid = MPI.COMM_WORLD.rank
smyid = '{:0>6d}'.format(myid)

m_ = 1.0
k_ = 1.0


def nicePrint(*s):
    MPI.COMM_WORLD.Barrier()
    for i in range(num_procs):
        MPI.COMM_WORLD.Barrier()
        if i == myid:
            print(str(myid)+': ' + ': '.join([str(ss) for ss in s]))
        MPI.COMM_WORLD.Barrier()


def run(order=1,
        prob=0,
        nsteps=100,
        dt=0.1,
        sc=1.0,
        visualization=False):

    class GradT(mfem.Operator):
        def __init__(self):
            mfem.Operator.__init__(self, 1)

        def Mult(self, x, y):
            y.Set(1.0/m_, x)

    class NegGradV(mfem.TimeDependentOperator):
        def __init__(self):
            mfem.TimeDependentOperator.__init__(self, 1)

        def Mult(self, x, y):
            if prob == 1:
                y[0] = - k_ * sin(x[0])
            elif prob == 2:
                y[0] = - k_ * x[0] * exp(-0.5 * x[0] * x[0])
            elif prob == 3:
                y[0] = - k_ * (1.0 + 2.0 * x[0] * x[0]) * x[0]
            elif prob == 4:
                y[0] = - k_ * (1.0 - 0.25 * x[0] * x[0]) * x[0]
            else:
                y[0] = - k_ * x[0]

    def hamiltonian(q, p, t):
        h = 1.0 - 0.5 / m_ + 0.5 * p * p / m_
        if prob == 1:
            h += k_ * (1.0 - cos(q))
        elif prob == 2:
            h += k_ * (1.0 - exp(-0.5 * q * q))
        elif prob == 3:
            h += 0.5 * k_ * (1.0 + q * q) * q * q
        elif prob == 4:
            h += 0.5 * k_ * (1.0 - 0.125 * q * q) * q * q
        else:
            h += 0.5 * k_ * q * q
        return h

    device = mfem.Device('cpu')
    if myid == 0:
        device.Print()

    # 2. Create and Initialize the Symplectic Integration Solver
    siaSolver = mfem.SIAVSolver(order)
    P = GradT()
    F = NegGradV()
    siaSolver.Init(P, F)

    # 3. Set the initial conditions
    t = 0.0
    q = mfem.Vector(1)
    p = mfem.Vector(1)
    e = mfem.Vector(nsteps+1)
    q[0] = sin(2*pi*myid/num_procs)
    p[0] = cos(2*pi*myid/num_procs)

    # 5. Create a Mesh for visualization in phase space
    nverts = 2*(nsteps+1)*num_procs if visualization else 0
    nelems = nsteps*num_procs if visualization else 0

    mesh = mfem.Mesh(2, nverts, nelems, 0, 3)

    part = mfem.intArray(nelems)

    # 6. Perform time-stepping
    e_mean = 0.0

    for i in range(nsteps):
        if i == 0:
            e[0] = hamiltonian(q[0], p[0], t)
            e_mean += e[0]
            if visualization:
                for j in range(num_procs):
                    mesh.AddVertex([0, 0, 0])
                    mesh.AddVertex([q[0], p[0], 0.0])

        #  6b. Advance the state of the system
        t, dt = siaSolver.Step(q, p, t, dt)
        e[i+1] = hamiltonian(q[0], p[0], t)
        e_mean += e[i+1]

        #  6d. Add results to GLVis visualization
        if visualization:
            for j in range(num_procs):
                mesh.AddVertex([0,  0, t])
                mesh.AddVertex([q[0], p[0], t])
                mesh.AddQuad([2*i*num_procs + 2*j,
                              2*(i+1)*num_procs + 2*j,
                              2*(i+1)*num_procs + 2*j+1,
                              2*i*num_procs + 2*j+1])
                part[num_procs*i + j] = j
            # this also works ;D
            # mesh.AddQuad(v.ToList())
            #mesh.AddQuad(np.array(v.ToList(), dtype=np.int32))
    #  7. Compute and display mean and standard deviation of the energy
    e_mean /= (nsteps + 1)
    e_var = 0.0
    for i in range(nsteps+1):
        e_var += (e[i] - e_mean)**2
    e_var /= (nsteps + 1)
    if myid == 0:
        print("Mean and standard deviation of the energy")
    nicePrint("{:g}".format(e_mean) + "\t" + "{:g}".format(sqrt(e_var)))

    #  9. Finalize the GLVis output
    if visualization:
        mesh.FinalizeQuadMesh(1)

        pmesh = mfem.ParMesh(MPI.COMM_WORLD, mesh, part.GetData())

        fec = mfem.H1_FECollection(1, 2)
        fespace = mfem.ParFiniteElementSpace(pmesh, fec)
        energy = mfem.ParGridFunction(fespace)
        energy.Assign(0.0)

        for i in range(nsteps+1):
            energy[2*i+0] = e[i]
            energy[2*i+1] = e[i]

        sock = mfem.socketstream("localhost", 19916)
        sock.precision(8)
        sock << "parallel " << num_procs << " " << myid << "\n"
        sock << "solution\n" << pmesh << energy
        sock << "window_title 'Energy in Phase Space'\n"
        sock << "keys\n maac\n" << "axis_labels 'q' 'p' 't'\n"
        sock.flush()


if __name__ == "__main__":
    from mfem.common.arg_parser import ArgParser

    parser = ArgParser(description='Ex20p (Sympletic ODE)')
    parser.add_argument('-m', '--mesh',
                        default='star.mesh',
                        action='store', type=str,
                        help='Mesh file to use.')
    parser.add_argument("-p",
                        "--problem-type",
                        action='store', type=int, default=0,
                        help=''.join(["Problem Type:\n",
                                      "\t  0 - Simple Harmonic Oscillator\n",
                                      "\t  1 - Pendulum\n",
                                      "\t  2 - Gaussian Potential Well\n",
                                      "\t  3 - Quartic Potential\n",
                                      "\t  4 - Negative Quartic Potential", ]))
    parser.add_argument('-o', '--order',
                        action='store', default=1, type=int,
                        help="Time integration order")
    parser.add_argument('-n', '--number-of-steps',
                        action='store', default=100, type=int,
                        help="Number of time steps")
    parser.add_argument('-dt', '--time-step',
                        action='store', default=0.1, type=float,
                        help="Time step size")
    parser.add_argument('-k', '--spring-constant',
                        action='store', default=1, type=float,
                        help="Sprint constant")
    parser.add_argument('-vis', '--visualization',
                        action='store_true',
                        default=True,
                        help='Enable GLVis visualization')
    parser.add_argument('-no-gp', '--no-gnuplot',
                        action='store_true',
                        default=True,
                        help='Disable GnuPlot visualization')

    args = parser.parse_args()
    if myid == 0:
        parser.print_options(args)

    prob = args.problem_type
    visualization = args.visualization
    order = args.order
    nsteps = args.number_of_steps
    dt = args.time_step
    sc = args.spring_constant
    np_gp = args.no_gnuplot

    run(order=order,
        prob=prob,
        nsteps=nsteps,
        dt=dt,
        sc=sc,
        visualization=visualization)
